#ifndef SCLINFO_H
#define SCLINFO_H

// ********** Class sClusterInfoEntry / sClusterInfo *****************************
template <class DATA>
class sClusterInfoEntry
{
public:
	sClusterInfoEntry(CLUSTERID Id) { id=Id; size=0; };
   virtual void write_info(ostream &os);							 	// save entry info

	CLUSTERID id;
	LONGINT	 size;
	CLUSTERID distanz;														// sort order field in SortedLinList
};

template <class DATA>
void sClusterInfoEntry<DATA>::write_info(ostream &os)
{
	os << "ClusterId: " << id
   	<< " Size: " << size
      << " (" << distanz << ")" << endl;
}

typedef sClusterInfoEntry<rstar_DATA> sClusterInfoEntryType;	// g++ workaround

// ********** Class sClusterInfo ***********************************************
template <class DATA>
class sClusterInfo : public clusterInfo<DATA>
{
public:
	sClusterInfo();
	virtual ~sClusterInfo();

   virtual void clear();
   virtual void createTemp(CLUSTERID Id);
	virtual void updateTemp(DATA *d);
   virtual void commitTemp();

 	// used by demo application
   virtual void 	  	write_info(char *fname, char *param_descr);	// save info (text)

   sClusterInfoEntryType *akt, *last;
	SortedLinList<sClusterInfoEntryType> *clInfoList;
};

template <class DATA>
sClusterInfo<DATA>::sClusterInfo()
{
	akt=last=NULL;
   clInfoList = new SortedLinList<sClusterInfoEntryType>();
	clInfoList->set_sorting(false);
}
template <class DATA>
sClusterInfo<DATA>::~sClusterInfo()
{
	delete akt;
   delete clInfoList;
}

template <class DATA>
void sClusterInfo<DATA>::clear()
{
	delete akt;
   delete clInfoList;
   clInfoList = new SortedLinList<sClusterInfoEntryType>;
	clInfoList->set_sorting(false);   
  	akt=last=NULL;
}
template <class DATA>
void sClusterInfo<DATA>::createTemp(CLUSTERID Id)
{
	if (akt) error("create: last entry not commited", true);
	akt = new sClusterInfoEntry<DATA>(Id);
}

template <class DATA>
void sClusterInfo<DATA>::updateTemp(DATA *d)
{	akt->size++; }                                   				// update size

template <class DATA>
void sClusterInfo<DATA>::commitTemp()
{
	akt->distanz=akt->size;													// default sort order is size
	clInfoList->insert(akt);
	last=akt; akt=NULL;
}

template <class DATA>
void sClusterInfo<DATA>::write_info(char *fname, char *param_descr)
{
	ofstream OutFile(fname);
  	if (OutFile==NULL) return;

	OutFile << param_descr << endl;
	for (sClusterInfoEntryType *it=clInfoList->get_first(); it; it=clInfoList->get_next())
   	it->write_info(OutFile);
}

#ifndef __BORLANDC__
template class sClusterInfo<rstar_DATA>;							// workaround for g++ template instantiation problem
#endif
#endif
