#ifndef __LINLIST
#define __LINLIST

#include <stdio.h>

////////////////////////////////////////////////////////////////////////
// LinList  (SLink)
////////////////////////////////////////////////////////////////////////

template <class DATA>
struct SLink
{
   DATA *d;                    	// Zeiger auf Element-Daten
   bool			doDelete;			// flag to delete DATA in destructor   
   SLink<DATA> *next;          	// Zeiger auf naechstes Element
   SLink<DATA> *prev;          	// Zeiger auf vorhergehendes Element

   SLink(bool _doDelete);
   ~SLink();
};

template <class DATA>
SLink<DATA>::SLink(bool _doDelete)
{
	d = NULL;
	next = prev = NULL;
   doDelete=_doDelete;
}

template <class DATA>
SLink<DATA>::~SLink()
{
	if (doDelete) delete d;
}

////////////////////////////////////////////////////////////////////////
// LinList
////////////////////////////////////////////////////////////////////////

template <class DATA>
class LinList
{
protected:
    SLink<DATA> *first;         // Rootzeiger des Datenbestands
    SLink<DATA> *last;          // Zeiger auf letztes Element
    int anz;                    // Anzahl der belegten Elemente in der Liste
    SLink<DATA> *akt;           // zeigt auf aktuelles Element
    int akt_index;              // Index des zuletzt mit get geholten Elements
    bool doDelete;				  // flag to delete DATA in destructor    
public:
    LinList(bool _doDelete=true);
    virtual ~LinList();
    virtual int get_num() 	     // gibt Anzahl der im Index belegten Elements
        { return anz; }         // zurueck

    virtual void check();       // ueberprueft Konsistenz der Liste

    virtual void insert(DATA *f);	// haengt ein Element vorne an die Liste an
	 virtual void noDelete();	  // DATA wird beim loeschen nicht beruecksichtigt
    virtual bool erase();       // loescht aktuelles Element aus der Liste

    virtual DATA * get(int i);  // liefert i-tes Element
    virtual DATA * get_first(); // liefert erstes Element im Index
    virtual DATA * get_last();  // liefert erstes Element im Index
    virtual DATA * get_next();  // liefert naechstes Element im Index
    virtual DATA * get_prev();  // liefert vorhergehendes Element im Index

    virtual void	insert_list(LinList *l);	// haengt ganze Liste l vorne an die Liste an, loescht l
};

template <class DATA>
void LinList<DATA>::insert_list(LinList *l)
{
   if (l == NULL) return;
   if (l->anz==0) return;

   // Zeiger umbiegen
   if (l->last != NULL)
	   l->last->next=first;
	if (first != NULL)
	   first->prev=l->last;

    // first, last und anz berichtigen
    anz+=l->anz;
    first=l->first;

    if (last == NULL)
	 last = l->last;

    // Position ist undefiniert
    akt = NULL;
    akt_index = -1;

    // l loeschen
    l->first=l->last=NULL;
    l->anz=0; delete l;
}

template <class DATA>
LinList<DATA>::LinList(bool _doDelete)
{
    anz = 0;
    akt_index = -1;
    akt = first = last = NULL;
    doDelete=_doDelete;
}

template <class DATA>
LinList<DATA>::~LinList()
{
	SLink<DATA> *hf;

   // Elemente freigeben
   akt = first;
   while (akt != NULL)
   {
		hf = akt->next;
		delete akt;
		akt = hf;
   }
}

template <class DATA>
void LinList<DATA>::insert(DATA *f)
{
    SLink<DATA> *sd;

    // neuen Rahmen erzeugen
    sd = new SLink<DATA>(doDelete);
    sd->d = f;

    // Zeiger umbiegen
    sd->next = first;
    sd->prev = NULL;
    if (first != NULL)
	first->prev = sd;

    // first, last und anz berichtigen
    anz++;
    first = sd;
    if (last == NULL)
	last = sd;

    // Position ist undefiniert
    akt = NULL;
    akt_index = -1;
}
template <class DATA>
void LinList<DATA>::noDelete()	  										// DATA wird beim loeschen nicht beruecksichtigt
{
	if (akt->doDelete) akt->doDelete=false;
 		else printf("LinList::noDelete: not owner of DATA");
}

template <class DATA>
bool LinList<DATA>::erase()
{
    SLink<DATA> *n_akt;

    // Liste leer oder akt nicht definiert?
    if (akt)
    {
	// Element ist erstes Element
	if (akt == first)
	{
	    // Element ist einziges Element
	    if (akt == last)
	    {
		akt_index = -1;
		first = last = NULL;
		n_akt = NULL;
	    }
	    else
	    {
		// Element ist erstes Element, aber nicht leztes
		(akt->next)->prev = NULL;
		first = akt->next;
		n_akt = first;
		akt_index = 0;
	    }
	}
	else
	{
	    // Element ist letztes Element
	    if (akt == last)
	    {
		(akt->prev)->next = NULL;
		last = akt->prev;
		n_akt = NULL;
		akt_index = -1;
	    }
	    else
	    // Element ist mitten in der Liste
	    {
		(akt->next)->prev = akt->prev;
		(akt->prev)->next = akt->next;
		n_akt = akt->next;
		akt_index++;
	    }
	}

	// Speicher freigeben
	delete akt;

	// aktuelles Element setzen
	akt = n_akt;

	// anz berichtigen
	anz--;
	return true;
    }
    return false;
}

template <class DATA>
DATA* LinList<DATA>::get(int i)
// liefert das i-te Element in der Liste
{
    bool ahead;   // wenn ahead true ist, wird in next-Richtung gesucht
    int j;

    // liegt das i-te Element ueberhaupt in der Liste?
    if (i >= anz)
	return NULL;

    // ist die Liste schon auf das i-te Element positioniert?
    if (i == akt_index)
	return akt->d;

    // hat eine Positionierung der Liste stattgefunden?
    if (akt_index == -1)
    {
	// i liegt naeher an first, als an last
	if (i < (anz / 2))
	{
	    akt = first;
	    akt_index = 0;
	    ahead = true;
	}
	else
	{
	    akt = last;
	    akt_index = anz - 1;
	    ahead = false;

	}
    }
    else
    {
	// die gewuenschte Position liegt vor der aktuellen
	if (i < akt_index)
	{
	    // liegt i naeher an first, als an akt_index?
	    if ((akt_index - i) > i)
	    {
		akt = first;
		akt_index = 0;
		ahead = true;
	    }
	    else
		ahead = false;
	}
	else
	{
	    // liegt i naeher an last, als an akt_index?
	    if ((i - akt_index) > ((anz-1) - i))
	    {
		akt = last;
		akt_index = anz - 1;
		ahead = false;
	    }
	    else
		ahead = true;
	}
    }


    // gesuchter Index liegt in next - Richtung
    if (ahead)
    {
	for (j = akt_index; j < i; j++)
	{
	    if (!akt)
		printf("LinList::get: List seems to be inkonsistent");
	    akt = akt->next;
	}
    }
    else
    {
	for (j = akt_index; j > i; j--)
	{
	    if (!akt)
		printf("LinList::get: List seems to be inkonsistent");
	    akt = akt->prev;
	}
    }

    akt_index = i;
    return akt->d;
}

template <class DATA>
DATA* LinList<DATA>::get_first()
{
    akt = first;

    if (akt != NULL)
    {
	akt_index = 0;
	return akt->d;
    }
    else
	return NULL;
}

template <class DATA> 
DATA* LinList<DATA>::get_last()
{
    akt = last;

    if (akt != NULL)
    {
	akt_index = anz - 1;
	return akt->d;
    }
    else
	return NULL;
}

template <class DATA> 
DATA* LinList<DATA>::get_next()
{
	if (akt==NULL) return NULL;
    akt = akt->next;
    if (akt != NULL)
    {
	akt_index++;
	return akt->d;
    }
    else
    {
	akt_index = -1;
	return NULL;
    }
}

template <class DATA>
DATA* LinList<DATA>::get_prev()
{
	if (akt==NULL) return NULL;
    akt = akt->prev;
    if (akt != NULL)
    {
	akt_index--;
	return akt->d;
    }
    else
    {
	akt_index = -1;
	return NULL;
    }
}

template <class DATA> 
void LinList<DATA>::check()
{
    SLink<DATA> *f, *old_f;
    int myanz;
    char buffer[255];

    old_f = first;
    // Liste muss ganz leer sein
    if (old_f == NULL)
    {
	if (last != NULL)
	    printf("LinList::check: first == NULL, last != NULL");
	if (anz != 0)
	    printf("LinList::check: first == NULL, anz != 0");
	return;
    }

    myanz = 1;
    if (old_f->prev != NULL)
    {
	printf("LinList::check: Listenkopf.prev ungleich NULL");
	return;
    }

    for (f = old_f->next; f != NULL; f = f->next)
    {
	if (f->prev != old_f)
	{
	    printf("LinList::check: Rueckwaertsverkettung fehlerhaft");
            return;
        }
	if (old_f->next != f)
	{
	    printf("LinList::check: Vorwaertsverkettung fehlerhaft");
            return;
        }
	old_f = f;

	myanz ++;
	if (myanz > anz)
	{
	    sprintf(buffer, "LinList::check: anz (%d != %d) passt nicht", myanz, anz);
	    printf(buffer);
	    return;
	}
    }

    if (old_f->next != NULL)
    {
	printf("LinList::check: Listenende.next ungleich NULL");
	return;
    }

    if (last != old_f)
    {
	printf("LinList::check: last ungleich Listenende");
	return;
    }

    if (myanz != anz)
    {
	sprintf(buffer, "LinList::check: anz (%d != %d) passt nicht", myanz, anz);
	printf(buffer);
    }
}


////////////////////////////////////////////////////////////////////////
// SortedLinList
////////////////////////////////////////////////////////////////////////

template <class DATA>
class SortedLinList : public LinList<DATA>
{
    bool increasing;
public:
    SortedLinList(bool _doDelete=true);

    virtual void set_sorting(bool _increasing); // wenn increasing gleich true (default), wird
                   						            // aufsteigend einsortiert
    virtual void insert(DATA *f);    			   // fuegt ein Element durch direktes Einfuegen ein
    virtual void sort(bool _increasing);			// sortiert die Liste
    virtual void check();               			// ueberprueft die Konsistenz der Liste
};

template <class DATA>
SortedLinList<DATA>::SortedLinList(bool _doDelete)
    : LinList<DATA>(_doDelete)
{
    increasing = true;
}


template <class DATA>
void SortedLinList<DATA>::set_sorting(bool _increasing)
{
    increasing = _increasing;
}

template <class DATA>
void SortedLinList<DATA>::check()
{
    SLink<DATA> *sd;
    DATA *old;
    bool valid;          // falls true, ist old mit einem Wert besetzt

    // Verkettung konsistent ?
    LinList<DATA>::check();

    // Sortierung korrekt ?
    valid = false;
    for (sd = this->first; sd != NULL; sd = sd->next)
    {
	 if (valid && (sd->d->distanz) < old->distanz)
         {
	     printf("LinList::check: Liste unsortiert");
	     return;
         }
	 valid = true;
	 old = sd->d;
    }
}

template <class DATA> 
void SortedLinList<DATA>::insert(DATA *f)
{
    SLink<DATA> *sd;

    // Rahmen fuer neues Datenelement erzeugen
    sd = new SLink<DATA>(this->doDelete);
    sd->d = f;

    // Liste leer?
    if (this->first == NULL)
    {
	this->first = sd;
	this->last = sd;
	this->anz = 1;
	sd->next = sd->prev = NULL;

	return;
    }

    // Einfuegestelle bestimmen
    if (increasing)
	for (this->akt = this->first; this->akt != NULL && (this->akt->d->distanz < f->distanz); this->akt = this->akt->next) ;
    else
	for (this->akt = this->first; this->akt != NULL && (this->akt->d->distanz > f->distanz); this->akt = this->akt->next) ;

    // neues Element muss vor this->akt eingefuegt werden --> Zeiger umbiegen
    if (this->akt != NULL)
    {
	if (this->akt == this->first)
	    this->first = sd;
	else
	    (this->akt->prev)->next = sd;
	sd->next = this->akt;
	sd->prev = this->akt->prev;
	this->akt->prev = sd;
    }
    else
    // neues Element muss als letztes Element eingefuegt werden -->
    // Zeiger umbiegen
    {
	sd->prev = this->last;
	sd->next = NULL;
	this->last->next = sd;
	this->last = sd;
    }

    // Gesamtanzahl erhoehen
    this->anz ++;
    this->akt_index = -1;
}

template <class DATA>
void SortedLinList<DATA>::sort(bool _increasing)
// Bubblesort der ganzen Liste
{
    SLink<DATA> *old_akt;
    DATA *s;
    bool any, swap;

    set_sorting(_increasing);

    any = true;
    while (any)
    {
	any = false;
	old_akt = NULL;
	for (this->akt = this->first; this->akt != NULL; this->akt = this->akt->next)
	{
	    // beim ersten Element gibts nichts zu vergleichen
	    if (old_akt != NULL)
	    {
		swap = false;
		if (_increasing)
		{
		    if ((this->akt->d->distanz) > (old_akt->d->distanz))
			swap = true;
		}
		else
		{
		    if ((this->akt->d->distanz) < (old_akt->d->distanz))
			swap = true;
		}

		if (swap)
		{
		    // es muessen nur die Datenpointer umgehaengt werden !
		    s = this->akt->d;
		    this->akt->d = old_akt->d;
		    old_akt->d = s;

		    // in diesem Durchlauf hat sich was getan --> 
		    // neuer Durchlauf
		    any = true;
		}
	    }
	    old_akt = this->akt;
	}    
    }

    this->akt_index = -1;
}

////////////////////////////////////////////////////////////////////////
// PersistentList
////////////////////////////////////////////////////////////////////////
// fuegt zur Klasse LinList noch Funktionen hinzu, die es erlauben,
// die Hauptspeicherliste auf Platte zu speichern und wieder zu lesen.
// Hierzu muss das Listenelement die Funktionen:
// 
// read_from_buffer 
// write_to_buffer 
// write_data_header
// 
// zur Verfuegung stellen. Diese Funktionen braucht die Liste, da sie
// ja keinerlei Informationen ueber die innere Struktur der Datenelemente
// hat. Ein einfaches Speichern des gesamten Speicherbereichs des Elements
// auf Platte ist wegen des VTable-Pointers nicht moeglich und wegen der 
// Moeglichkeit, dass im Objekt Verweise auf andere Objekte gespeichert 
// sind, die bei einem erneuten Programmlauf und dem Laden der Liste
// keinen Sinn mehr machen wuerden. Um das Element beim Laden richtig 
// erzeugen zu koennen, steht vor jedem Element im File ein Header, 
// dessen Laenge ueber die virtuelle Funktion "get_data_header_size" 
// gesetzt werden kann. Ebenso steht am Anfang des Files ein 
// Header, der beliebig beschrieben werden kann. Um diese Features
// nutzen zu koennen, muss eine von PersistentList abgeleitete Klasse
// die folgenden Funktionen zur Verfuegung stellen:
// 
// get_header_size 
// get_data_header_size
// read_header
// write_header
// 
// Beim Lesen und Schreiben der Liste von Platte ruft PersistentList 
// die Funktion read/write_header mit einem Zeiger auf einen Puffer 
// auf. Die abgeleitete Klasse muss dann aus diesem Puffer ihre Daten 
// lesen bzw. schreiben. 
// Alle Objekte, die in diesem Container gespeichert werden, muessen
// einen Konstruktor KONSTRUKTOR(void *) zur Verfuegung stellen.
// Diesem werden die Headerdaten des einzelnen Elements uebergeben.
// Das File hat damit folgendes Format:
//
//  Byte           Bezeichnung             Inhalt
//  sizeof(int)    Headergroesse             h
//  sizeof(int)    Anzahl d. Elemente        n   
//  sizeof(int)    Datenheadergroesse       dh    
//  h              Header
//
//  n-mal 
//        sizeof(int)    Datengroesse             d
//        dh             Datenheader    
//        d              Daten
////////////////////////////////////////////////////////////////////////

template <class DATA>
class PersistentList : public LinList<DATA>
{
protected:
    virtual int get_header_size();            // liefert Groesse des Headers
    virtual int get_data_header_size();       // liefert Groesse des Elementheaders
    virtual void read_header(void *buffer);   // liest Header in Objekt ein
    virtual void write_header(void *buffer);  // liest Header in Objekt ein
public:
    void save(char *name);      // speichert ganze Liste auf Platte
    void load(char *name);      // liest ganze Liste von Platte
};


template <class DATA> 
int PersistentList<DATA>::get_header_size()
{
    return 0;
}

template <class DATA> 
int PersistentList<DATA>::get_data_header_size()
{
    return 0;
}

template <class DATA> 
void PersistentList<DATA>::read_header(void *buffer)
{
}

template <class DATA> 
void PersistentList<DATA>::write_header(void *buffer)
{
}

template <class DATA> 
void PersistentList<DATA>::load(char *name)
{
    FILE *fp;
    DATA *d;
    void *buffer, *d_buffer, *header, *d_header;
    int header_size, num_elem, d_header_size, d_size;
    int i;

    // File oeffnen
    fp = fopen(name, "rb");
    if (!fp)
	return;

    // Headergroesse lesen
    if (fread(&header_size, sizeof(int), 1, fp) != 1)
	printf("PersistentList::load: cannot read size of header from file");

    // Elementzahl lesen
    if (fread(&num_elem, sizeof(int), 1, fp) != 1)
	printf("PersistentList::load: cannot read number of elements from file");

    // Datenheadergroesse lesen
    if (fread(&d_header_size, sizeof(int), 1, fp) != 1)
	printf("PersistentList::load: cannot read size of data header from file");

    // Header allokieren und einlesen
    header = new char[header_size];
    if (header_size > 0 && fread(header, header_size, 1, fp) != 1)
	printf("PersistentList::save: cannot read  header from  file");
    read_header(header);

    // Datenheader allokieren
    d_header = new char[d_header_size];

    for (i = 0; i < num_elem; i++)
    {
	// Laenge des Elements einlesen
	if (fread(&d_size, sizeof(int), 1, fp) != 1)
	    printf("PersistentList::load: cannot read element size from file");

	// Datenheader einlesen
	if (d_header_size > 0 && fread(d_header, d_header_size, 1, fp) != 1)
	    printf("PersistentList::load: cannot read data header from file");

	// neues Element erzeugen
	d = new DATA(d_header);

	// Daten einlesen
	buffer = new char[d_size];
	if (d_size > 0 && fread(buffer, d_size, 1, fp) != 1)
	    printf("PersistentList::load: cannot read data from file");
	d->read_from_buffer(buffer);
	delete [] buffer;

	// Element in die Liste einfuegen
	insert(d);
    }

    // Puffer freigeben
    delete [] header;
    delete [] d_header;

    // File schliessen
    fclose(fp);
}

template <class DATA> 
void PersistentList<DATA>::save(char *name)
{
    FILE *fp;
    DATA *d;
    void *buffer, *d_buffer, *header, *d_header;
    int header_size, num_elem, d_header_size, d_size;
    int i;

    // File oeffnen
    fp = fopen(name, "wb");
    if (!fp)
	printf("PersistentList::save: cannot open file for writing");

    // Headergroesse schreiben
    header_size = get_header_size();
    if (fwrite(&header_size, sizeof(int), 1, fp) != 1)
	printf("PersistentList::save: cannot write size of header to file");

    // Elementzahl schreiben
    num_elem = this->get_num();
    if (fwrite(&num_elem, sizeof(int), 1, fp) != 1)
	printf("PersistentList::save: cannot write number of elements to file");

    // Datenheadergroesse schreiben
    d_header_size = get_data_header_size();
    if (fwrite(&d_header_size, sizeof(int), 1, fp) != 1)
	printf("PersistentList::save: cannot write size of data header to  file");

    // Header allokieren schreiben
    header = new char[header_size];
    write_header(header);
    if (header_size > 0 && fwrite(header, header_size, 1, fp) != 1)
	printf("PersistentList::save: cannot write header to  file");

    // Datenheader allokieren
    d_header = new char[d_header_size];

    for (d = this->get_first(); d != NULL; d = this->get_next())
    {
	// Laenge des Elements schreiben
	d_size = d->get_size();
	if (fwrite(&d_size, sizeof(int), 1, fp) != 1)
	    printf("PersistentList::save: cannot write element size to file");

	// Datenheader schreiben
	d->write_data_header(d_header);
	if (d_header_size > 0 && fwrite(d_header, d_header_size, 1, fp) != 1)
	    printf("PersistentList::save: cannot write data header to file");

	// Daten schreiben
	buffer = new char[d_size];
	d->write_to_buffer(buffer);
	if (d_size > 0 && fwrite(buffer, d_size, 1, fp) != 1)
	    printf("PersistentList::save: cannot read data from file");
	delete [] buffer;
    }

    // Puffer freigeben
    delete [] header;
    delete [] d_header;

    // File schliessen
    fclose(fp);
}


#endif  // __LINLIST
