#ifndef rstar_DATA_H
#define rstar_DATA_H

#include <limits.h>
#include <float.h>

extern void error(char *t, bool ex);
extern int RTDataNode__dimension;

class rstar_DATA
{
//_________________________ compulsory part ____________________________
public:
   int dimension;
   float *data;																// Vector
   float distanz;																// distance from NN(rstar_DATA)
   float *get_mbr();															// returns MBR of the object
   float get_area()															// returns the area of the MBR
   { return 0.0; }															// for vectors always 0.0
   void read_from_buffer(char *buffer);								// reads data from buffer
   void write_to_buffer(char *buffer);									// writes data to buffer
   int get_size();															// returns amount of needed space
   rstar_DATA & operator = (rstar_DATA &_d);
   int compare(rstar_DATA *d);

// rstar_DATA (const rstar_DATA &_d);									// copy constructor (required in some cases)
   rstar_DATA(int *dimension = &RTDataNode__dimension);
   ~rstar_DATA();
//________________________ optional part ________________________________
	LONGINT 		id; 		                                      	// Object Id
	CLUSTERID  	clusterId;
	float			weight;					 									// point weight
	int	rank;
	bool	printed;
	int	y;
};

// ********** rstar_DATA implementation ****************************************
rstar_DATA::rstar_DATA(int *_dimension)
{
    dimension = *((int *)_dimension);
    data = new float[dimension];
    printed = false;
    rank = -1;
    y = -1;
}

/*rstar_DATA::rstar_DATA(const rstar_DATA &_d)							// copy constructor (required in some cases)
{
   dimension = _d.dimension;
   data = new float[dimension];
   distanz = _d.distanz;
   clusterId  = _d.clusterId;
   weight = _d.weight;
   id = _d.id;
   memcpy(data, _d.data, sizeof(float) * dimension);
}*/

rstar_DATA::~rstar_DATA()
{
	delete[] data;
}

rstar_DATA & rstar_DATA::operator = (rstar_DATA &_d)
{
   dimension = _d.dimension;
   distanz = _d.distanz;
   clusterId  = _d.clusterId;
   id = _d.id;
   weight = _d.weight;
   memcpy(data, _d.data, sizeof(float) * dimension);
   rank = _d.rank;
   printed = _d.printed;
   y = _d.y;
   return *this;
}

void rstar_DATA::read_from_buffer(char *buffer)
{
   int i;
   i = dimension*sizeof(float);
   memcpy(data, buffer, i);

   memcpy(&distanz, &buffer[i], sizeof(float));
   i += sizeof(float);
   memcpy(&clusterId, &buffer[i], sizeof(int));
   i += sizeof(CLUSTERID);
   memcpy(&id, &buffer[i], sizeof(LONGINT));
   i += sizeof(LONGINT);
   memcpy(&weight, &buffer[i], sizeof(weight));
   i += sizeof(float);
   memcpy(&rank, &buffer[i], sizeof(rank));
   i += sizeof(int);
   memcpy(&printed, &buffer[i], sizeof(printed));
   i += sizeof(bool);
   memcpy(&y, &buffer[i], sizeof(y));
   i += sizeof(int);
}

void rstar_DATA::write_to_buffer(char *buffer)
{
   int i;
   i = dimension*sizeof(float);
   memcpy(buffer, data, i);

   memcpy(&buffer[i], &distanz, sizeof(float));
   i += sizeof(float);
   memcpy(&buffer[i], &clusterId, sizeof(int));
   i += sizeof(CLUSTERID);
   memcpy(&buffer[i], &id, sizeof(LONGINT));
   i += sizeof(LONGINT);
   memcpy(&buffer[i], &weight, sizeof(float));
   i += sizeof(float);
   memcpy(&buffer[i], &rank, sizeof(int));
   i += sizeof(int);
   memcpy(&buffer[i], &printed, sizeof(bool));
   i += sizeof(bool);
   memcpy(&buffer[i], &y, sizeof(int));
   i += sizeof(int);
}

int rstar_DATA::get_size()
{   // data + distanz + clusterId + id + weight
    return dimension * sizeof(float)
           + sizeof(float)
           + sizeof(CLUSTERID)
           + sizeof(LONGINT)
           + sizeof(float)
	   + sizeof(int)
	   + sizeof(bool)
	   + sizeof(int);
}

float *rstar_DATA::get_mbr()
{	// point data: upper = lower boundary
	float *f = new float[2*dimension];
   for (int i = 0; i < dimension; i++)
		f[2*i] = f[2*i+1] = data[i];
   return f;
}

int rstar_DATA::compare(rstar_DATA *d)
{
   for(int i=0; i<dimension; i++)
   {
      if (d->data[i] < data[i]) return -1;
      if (d->data[i] > data[i]) return  1;
	}
	return 0;
}
#endif /* rstar_DATA_H */
