#ifndef __poly_RTREE
#define __poly_RTREE

#include "linlist.h"
#include "blk_file.h"		

#include <limits.h>
#include <float.h>
#include <values.h>

//////////////////////////////////////////////////////////////////////////////
// defines
//////////////////////////////////////////////////////////////////////////////
#define MAXREAL 9.99e20
#define PTS 6

#define mini(a, b) (((a) < (b))? (a) : (b)  )
#define maxi(a, b) (((a) > (b))? (a) : (b)  )
//////////////////////////////////////////////////////////////////////////////
// C-prototypes
//////////////////////////////////////////////////////////////////////////////
float area(int dimension, float *mbr);
float margin(int dimension, float *mbr);
float overlap(int dimension, float *r1, float *r2);
float objectDIST(const void  *Point1,const void  *Point2);
float MINMAXDIST(const void  *Point, float *bounces);
float MINDIST(const void *Point, float *bounces);

bool inside(float &p, float &lb, float &ub);
void enlarge(int dimension, float **mbr, float *r1, float *r2);
bool section(int dimension, float *mbr1, float *mbr2);
bool section_c(int dimension, float *mbr1, const void *center, float radius);

void check_mbr(int dimension, float *mbr);

int sort_lower_mbr(const void *d1, const void *d2);
int sort_upper_mbr(const void *d1, const void *d2);
int sort_center_mbr(const void *d1, const void *d2);
int sortmindist(const void *element1, const void *element2);
int pruneBrunchList(float &nearest_distanz, const void *activebrunchList, int n, int downward = true);
int testBrunchList(const void *abL, const void *sL, int n, int last, int downward = true);

void error(char *t, bool ex);
//////////////////////////////////////////////////////////////////////////////
// externals
//////////////////////////////////////////////////////////////////////////////
extern int RTDirNode__dimension;
extern void* RTDirNode__my_tree;
extern int RTDataNode__dimension;
#ifdef ZAEHLER
extern float page_access;
#endif
//////////////////////////////////////////////////////////////////////////////
// enums / types
//////////////////////////////////////////////////////////////////////////////
enum SECTION {OVERLAP, INSIDE, S_NONE};
enum R_OVERFLOW {SPLIT, REINSERT, NONE};
typedef float *floatptr;

//////////////////////////////////////////////////////////////////////////////
// classes
//////////////////////////////////////////////////////////////////////////////
template <class DATA> class RTree;			// AAA
template <class DATA> class DirEntry;     // AAA

struct SortMbr
{
    int dimension;
    float *mbr;
    float *center;
    int index;
};

// only of rtree_data contains a virtual function (destructor..):
/*#ifdef __BORLANDC__															// AAA
struct DataStruct
{
	 int dummy;
    int dimension;
    float *Data;
};
#else              																// AAA*/
struct DataStruct
{
    int dimension;
    float *Data;
};
//#endif             	 														// AAA

struct BranchList
{
    int entry_number;
    float mindist;
    float minmaxdist;
    bool section;
};

template <class DATA>
class RTNode
{
    friend class RTree<DATA>;
protected:
    RTree<DATA> *my_tree;               // pointer to R-tree
    int capacity;                       // max. # of entries
    int dimension;
    int num_entries;                    // # of used entries
    bool dirty;                         // true, if node has to be written
public:
    int block;                          // disc block
    char level;                         // level of the node in the tree

    int get_num()                       // returns # of used entries
    { return num_entries;}
    virtual int get_num_of_data() = 0;  // returns number of data entries
                                        // behind that node

    virtual DATA * get(int i, int dcopy = true) = 0;      
                                        // returns the i-th object in the 
                                        // tree lying behind that node
    
    virtual void read_from_buffer(char *buffer) = 0;
                                        // reads data from buffer
    virtual void write_to_buffer(char *buffer) = 0;
                                        // writes data to buffer

    virtual bool is_data_node() = 0;	// returns true, if "this" is 
                                        // *RTDataNode 

    virtual float *get_mbr() = 0;       // returns mbr enclosing whole page
    virtual void print()                // prints rectangles
    { }

    int split(float **mbr, int **distribution);
                                        // splits an Array of mbr's into 2 
                                        // and returns in distribution
                                        // which *m mbr's are to be
                                        // moved
    virtual void NearestNeighborSearch(DATA *Point, DATA  *&Nearest,float &nearest_distanz) = 0;

    virtual void NearestNeighborSearch(DATA *Point, 
		SortedLinList<DATA> *&res, float &nearest_distanz) =  0;

    virtual void rangeQuery(float *mbr, SortedLinList<DATA> *res) = 0;

    virtual void rangeQuery(float *mbr, LinList<DATA> *res) = 0;

    virtual void rangeQuery(DATA *center, float radius, SortedLinList<DATA> *res) = 0; 

    virtual void range_nnQuery(float *mbr,SortedLinList<DATA> *res, 
				DATA *center, float &nearest_distanz,
				DATA *Nearest, bool *success) = 0;

    virtual void overlapping(float *p, int *nodes_t) = 0;
    virtual void nodes(int *nodes_a) = 0;

    virtual void writeinfo(FILE *f) = 0;

    virtual void get_dmbr(float* dmbr, int& length) = 0;
					//get mbr for data pages; 
	
    virtual R_OVERFLOW insert(DATA *d, RTNode<DATA> **sn) = 0; 
                                        // inserts d recursivly, if there
                                        // occurs a split, false will be
                                        // returned and in sn a 
                                        // pointer to the new node

    virtual void region(float *mbr) = 0;
                                        // prints all entries sectioning
                                        // mbr
 
    RTNode(RTree<DATA> *rt);
    RTNode(RTree<DATA> *rt, int _block);
    virtual ~RTNode();
};

template <class DATA> 
class RTDirNode: public RTNode<DATA>
{
    friend class RTree<DATA>;

    DirEntry<DATA> *entries;            // array of entries
    bool son_is_data;                   // true, if son is a data page
public:
    void read_from_buffer(char *buffer);// reads data from buffer
    void write_to_buffer(char *buffer); // writes data to buffer

    bool is_data_node()
    {return false;}

    int get_num_of_data();              // returns number of data entries
                                        // behind that node

    DATA * get(int i, int dcopy = true);// returns the i-th object in the 
                                        // tree lying behind that node

    void print();                       // prints rectangles

    float *get_mbr();                   // returns mbr enclosing whole page

    void enter(DirEntry<DATA> *de);     // inserts new entry

    void split(RTDirNode<DATA> *sn);    // splits directory page to *this and sn

    int choose_subtree(float *brm);     // chooses best subtree for insertion

    void NearestNeighborSearch(DATA *QueryPoint, 
				DATA *&Nearest,float &nearest_distanz);

    void NearestNeighborSearch(DATA *QueryPoint, 
			SortedLinList<DATA> *&res,
			float &nearest_distanz);

    void rangeQuery(float *mbr, SortedLinList<DATA> *res);

    void rangeQuery(float *mbr, LinList<DATA> *res);

    void rangeQuery(DATA *center, float radius, SortedLinList<DATA> *res);	

    void range_nnQuery(float *mbr, SortedLinList<DATA> *res,
			DATA *center, float &nearest_distanz, DATA *Nearest,
			bool *success);

    R_OVERFLOW insert(DATA *d, RTNode<DATA> **sn);  
                                        // inserts d recursivly, if there
                                        // occurs a split, false will be
                                        // returned and in sn a 
                                        // pointer to the new node
    void region(float *mbr);            // prints all entries sectioning
                                        // mbr
    void overlapping(float *p, int *nodes_t);
    void nodes(int *nodes_a);
    void writeinfo(FILE  *f);
    void get_dmbr(float* dmbr,int& length);	// get mbr for data pages;

    RTDirNode(RTree<DATA> *rt);                  
    RTDirNode(RTree<DATA> *rt, int _block);
    virtual ~RTDirNode();
    int kNN_pruneBrunchList(float &nearest_distanz, const void *activebrunchList, SortedLinList<DATA> *res, int n, int downward = true);
};

template <class DATA> 
class RTDataNode: public RTNode<DATA>
{
    DATA *data;                         // array of data
public:
    void read_from_buffer(char *buffer);// reads data from buffer
    void write_to_buffer(char *buffer); // writes data to buffer

    bool is_data_node()
    {return true;}

    void print();                       // prints vector

    int get_num_of_data();              // returns numbre of data entries
                                        // in that node
    DATA * get(int i, int dcopy = true);// returns the i-th object in this node 

    float *get_mbr();                   // returns mbr enclosing whole page 

    void split(RTDataNode<DATA> *sn);   // splits data page into sn
                                        // *this  
    void NearestNeighborSearch(DATA *QueryPoint, 
				DATA *&Nearest, float &nearest_distanz);

    void NearestNeighborSearch(DATA *QueryPoint, 
			SortedLinList<DATA> *&res, float &nearest_distanz);

    void rangeQuery(float *mbr, SortedLinList<DATA> *res);

    void rangeQuery(float *mbr, LinList<DATA> *res);

    void rangeQuery(DATA *center, float radius, SortedLinList<DATA> *res);

    void range_nnQuery(float *mbr, SortedLinList<DATA> *res,
			DATA *center, float &nearest_distanz,
			DATA *Nearest, bool *success);
  
    R_OVERFLOW insert(DATA *d, RTNode<DATA> **sn);  
                                        // inserts d recursivly, if there
                                        // occurs a split, false will be
                                        // returned and in sn a 
                                        // pointer to the new node

    void region(float *mbr);            // prints all entries sectioning
                                        // mbr

    void overlapping(float *p, int *nodes_t)
    { nodes_t[0]++; }

    void nodes(int *nodes_a)
    { nodes_a[0]++; }

    void writeinfo(FILE *f);
    void get_dmbr(float* dmbr,int& length);	// get mbr for data pages;
    RTDataNode(RTree<DATA> *rt);                  
    RTDataNode(RTree<DATA> *rt, int _block);
    virtual ~RTDataNode();
};


template <class DATA> 
class DirEntry
{
    friend class RTDirNode<DATA>;
    friend class RTree<DATA>;

    RTree<DATA> *my_tree;               // pointer to my R-tree
    int son;                            // block # of son
    RTNode<DATA> *son_ptr;              // pointer to son if in main mem.
    bool son_is_data;                   // true, if son is a data page
    int dimension;                      // dimension of the box
    float *bounces;                     // pointer to box
    int son_level;                      // level of the node pointed to
    int num_of_data;                    // amount of data entries behind the
                                        // son of this entry
public:
    bool is_inside(float *v);           // tests, if v is inside the box
    SECTION section(float *mbr);        // tests, if mbr intersects the box
    RTNode<DATA>* get_son();            // returns pointer to the son,
                                        // loads son if necessary

    bool section_circle(DATA *center, float radius);

    void read_from_buffer(char *buffer);// reads data from buffer   
    void write_to_buffer(char *buffer); // writes data to buffer  

    int get_size();                     // returns amount of needed buffer space

    virtual DirEntry<DATA> & operator = (DirEntry<DATA> &_d);

    DirEntry(int dimension = RTDirNode__dimension, 
	     RTree<DATA> *rt = (RTree<DATA> *)RTDirNode__my_tree);
    virtual ~DirEntry();
};

template <class DATA> 
class RTQuery
{
protected:
    LinList<DATA> dlist;
public:
    virtual bool intersect(float *mbr) = 0;
    virtual bool intersect(DATA *d) = 0;
};


template <class DATA> 
class RTree 
{
public:
    friend class RTNode<DATA>;
    friend class RTDirNode<DATA>;
    friend class RTDataNode<DATA>;

    int root;                            // block # of root node
    RTNode<DATA> *root_ptr;              // root-node
    bool root_is_data;                   // true, if root is a data page
    int dimension;                       // dimension of the data's

    int num_of_data;	                 // # of stored data
    int num_of_dnodes;	                 // # of stored data pages
    int num_of_inodes;	                 // # of stored directory pages

    bool *re_level;                      // if re_level[i] is true,
                                         // there was a reinsert on level i
    LinList<DATA> re_data_cands;         // data entries to reinsert

    CachedBlockFile *file;	         // storage manager for harddisc blocks

    void load_root();                    // loads root_node into memory
    int akt;                             // # of actually got data (get_next)
    char *header;
    float node_weight[20];               // weight for simulation of cache
protected:
    char *user_header;
    void read_header(char *buffer);      // reads Rtree header
    void write_header(char *buffer);     // writes Rtree header
public:
    int get_num()                        // returns # of stored data
    { return num_of_data; }              

    void insert(DATA *d);                // inserts new data into tree

    DATA *get_first()                    // trace
    { akt = 0; return get(akt); }        // through
    DATA *get_next()                     // all 
    { akt++; return get(akt); }          // data

    DATA *get(int i, int dcopy = true);  // get i-th data

    DATA *get_upd(int i){get(i, false);} // get i-th data for Update

    void set_node_weight(int cache_size); // calculate influence of cache

    void region(float *mbr);             // print all entries intersection 
                                         // the box mbr

    void sortBranchList( LinList<DATA>  *branchList);
    
    void NearestNeighborQuery(DATA *QueryPoint, DATA *&Nearest);
    
    void NearestNeighborQuery(DATA *QueryPoint, SortedLinList<DATA> *&res);

    void k_NearestNeighborQuery(DATA *QueryPoint, int k, 
				SortedLinList<DATA> *&NeighborList);
    void rangeQuery(float *mbr, SortedLinList<DATA> *res);

    void rangeQuery(float *mbr, LinList<DATA> *res);

    void rangeQuery(DATA *center, float radius, SortedLinList<DATA> *res);

    void range_nnQuery(float *mbr, SortedLinList<DATA> *res, DATA *Nearest);

    void SampleQuery(float SampleSizePerCent, LinList<DATA> *res);

    void SeedsQuery(LinList<DATA> *samples, SortedLinList<DATA> *res);

    void printinfo(FILE *f);

    void overlapping(float *p, int *nodes_t);
    void nodes(int *nodes_a);

    void writeinfo(FILE *f);

    void get_dmbr(float* dmbr,int& length);	// get mbr for data pages;

    RTree(char *fname,
	  int _b_length, int cache_size,
	  int _dimension);
    RTree(char *fname, int cache_size);

    bool prototype;
    RTree() { prototype=true; } 		// allow creation of "cloneable" descendent

    virtual ~RTree();
};


//////////////////////////////////////////////////////////////////////////////
// implementation
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// DirEntry
//////////////////////////////////////////////////////////////////////////////

template <class DATA> 
DirEntry<DATA>::DirEntry(int _dimension, RTree<DATA> *rt)
{
    dimension = _dimension;
    my_tree = rt;
    bounces = new float[2*dimension];
    son_ptr = NULL;
    num_of_data = 0;
}

template <class DATA> 
DirEntry<DATA>::~DirEntry()
{
    delete [] bounces;
    if (son_ptr != NULL)
	delete son_ptr;
}

template <class DATA> 
DirEntry<DATA>& DirEntry<DATA>::operator = (DirEntry &_d)
{
    dimension = _d.dimension;
    son = _d.son;
    son_ptr = _d.son_ptr;
    son_level = _d.son_level;
    son_is_data = _d.son_is_data;
    num_of_data = _d.num_of_data;
    memcpy(bounces, _d.bounces, sizeof(float) * 2 * dimension);

    return *this;
}

template <class DATA> 
bool DirEntry<DATA>::is_inside(float *v)
{
    int i;

    for (i = 0; i < dimension; i++)
    {
	if (v[i] < bounces[2*i] ||     // untere Grenze
	    v[i] > bounces[2*i + 1])   // obere Grenze
	    return false;
    }

    return true;
}

template <class DATA> 
SECTION DirEntry<DATA>::section(float *mbr)
// testet, ob mbr den Eintrag irgendwo ueberlappt
// Uberlappung kann nur dann nicht auf, wenn 
//
//    ttttt eeeee
//
//      oder
//
//    eeeee ttttt
//
// diese beiden Faelle werden fuer jede Dimension getestet
{
    bool inside;
    bool overlap;
    int i;

    overlap = true;
    inside = true;
    for (i = 0; i < dimension; i++)
    {
	if (mbr[2*i]     > bounces[2*i + 1] ||
	    mbr[2*i + 1] < bounces[2*i]) 
	    overlap = false;
	if (mbr[2*i]     < bounces[2*i] ||
	    mbr[2*i + 1] > bounces[2*i + 1]) 
	    inside = false;
    }
    if (inside)
	return INSIDE;
    else if (overlap)
	return OVERLAP;
    else
	return S_NONE;
}


template <class DATA> 
void DirEntry<DATA>::read_from_buffer(char *buffer)
{
    int i;

    i = 2*dimension*sizeof(float);
    memcpy(bounces, buffer, i);
    memcpy(&son, &buffer[i], sizeof(int));
    i += sizeof(int);
    memcpy(&num_of_data, &buffer[i], sizeof(int));
}

template <class DATA> 
void DirEntry<DATA>::write_to_buffer(char *buffer)
{
    int i;

    i = 2*dimension*sizeof(float);
    memcpy(buffer, bounces, i);
    memcpy(&buffer[i], &son, sizeof(int));
    i += sizeof(int);
    memcpy(&buffer[i], &num_of_data, sizeof(int));
}

template <class DATA> 
int DirEntry<DATA>::get_size()
{
    return 2*dimension * sizeof(float) + sizeof(int) + sizeof(int);
}

template <class DATA> 
RTNode<DATA>* DirEntry<DATA>::get_son()
{


    if (son_ptr == NULL)
    {
	if (son_is_data)
	    son_ptr = new RTDataNode<DATA>(my_tree, son);
	else
	    son_ptr = new RTDirNode<DATA>(my_tree, son);
    }
    return son_ptr;
}

template <class DATA>
bool DirEntry<DATA>::section_circle(DATA *center, float radius)
{
 float r2;

 r2 = radius * radius;

 if ( (r2 - MINDIST(center,bounces )) >= 1.0e-6)
	return true;
 else
	return false;
}
 


//////////////////////////////////////////////////////////////////////////////
// RTNode
//////////////////////////////////////////////////////////////////////////////

template <class DATA> 
RTNode<DATA>::~RTNode()
{
}

template <class DATA> 
RTNode<DATA>::RTNode(RTree<DATA> *rt)
{
    my_tree = rt; 
    dimension = rt->dimension; 
    num_entries = 0; 

    // noch kein Plattenblock vergeben
    block = -1;
}

template <class DATA> 
int RTNode<DATA>::split(float **mbr, int **distribution)
// teilt ein Array von mbr's in zwei Haelften und liefert in distribution
// zurueck, welche *m Eintraege in die neue Seite verschoben werden muessen
// hierfuer wird der R*-Baum Split Algorithmus benutzt
{
#ifdef SHOWMBR
	split_000++;
#endif

    bool lu;
    int i, j, k, l, s, n, m1, dist, split_axis;
    SortMbr *sml, *smu;
    float minmarg, marg, minover, mindead, dead, over, 
	*rxmbr, *rymbr;

    // how much nodes are used?
    n = get_num();

    // nodes have to be filled at least 40%
    m1 = (int) ((float)n * 0.40); 

    // sort by lower value of their rectangles
    // Indexarray aufbauen und initialisieren
    sml = new SortMbr[n];    
    smu = new SortMbr[n];    
    rxmbr = new float[2*dimension];
    rymbr = new float[2*dimension];

    // choose split axis
    minmarg = MAXREAL;
    for (i = 0; i < dimension; i++)
    // for each axis
    {
        for (j = 0; j < n; j++)
        {
            sml[j].index = smu[j].index = j;
            sml[j].dimension = smu[j].dimension = i;
            sml[j].mbr = smu[j].mbr = mbr[j];
        }

        // Sort by lower and upper value perpendicular axis_i
      	qsort(sml, n, sizeof(SortMbr), sort_lower_mbr);
        qsort(smu, n, sizeof(SortMbr), sort_upper_mbr);

        marg = 0.0;
        // for all possible distributions of sml
        for (k = 0; k < n - 2*m1 + 1; k++)
        {
            // now calculate margin of R1
	    // initialize mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for (l = 0; l < m1+k; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =   mini(rxmbr[s],   sml[l].mbr[s]);
	            rxmbr[s+1] = maxi(rxmbr[s+1], sml[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 

            // now calculate margin of R2
	    // initialize mbr of R2 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for ( ; l < n; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =   mini(rxmbr[s],   sml[l].mbr[s]);
	            rxmbr[s+1] = maxi(rxmbr[s+1], sml[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 
        }

        // for all possible distributions of smu
       	for (k = 0; k < n - 2*m1 + 1; k++)
        {
            // now calculate margin of R1
	    // initialize mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for (l = 0; l < m1+k; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =  mini(rxmbr[s],   smu[l].mbr[s]);
	            rxmbr[s+1] = maxi(rxmbr[s+1], smu[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 

            // now calculate margin of R2
	    // initialize mbr of R2 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for ( ; l < n; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =   mini(rxmbr[s],   smu[l].mbr[s]);
	            rxmbr[s+1] = maxi(rxmbr[s+1], smu[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 
        }

        // actual margin better than optimum?
        if (marg < minmarg)
        {
            split_axis = i;
            minmarg = marg;
        }

    }

    /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
//    printf("split axis = %d\n", split_axis);
	       

    // choose best distribution for split axis
    for (j = 0; j < n; j++)
    {
	sml[j].index = smu[j].index = j;
	sml[j].dimension = smu[j].dimension = split_axis;
	sml[j].mbr = smu[j].mbr = mbr[j];
    }
	       
    // Sort by lower and upper value perpendicular split axis
    qsort(sml, n, sizeof(SortMbr), sort_lower_mbr);
    qsort(smu, n, sizeof(SortMbr), sort_upper_mbr);
    
    minover = MAXREAL;
    mindead = MAXREAL;
    // for all possible distributions of sml and snu
    for (k = 0; k < n - 2*m1 + 1; k++)
    {
        // lower sort
	// now calculate margin of R1
	// initialize mbr of R1 
        dead = 0.0;
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rxmbr[s] =    MAXREAL;
	    rxmbr[s+1] = -MAXREAL;
	}
	for (l = 0; l < m1+k; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =   mini(rxmbr[s],   sml[l].mbr[s]);
		rxmbr[s+1] = maxi(rxmbr[s+1], sml[l].mbr[s+1]);
	    }
            dead -= area(dimension, sml[l].mbr);
	}
        dead += area(dimension, rxmbr);
	
	// now calculate margin of R2
	// initialize mbr of R2 
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rymbr[s] =    MAXREAL;
       	    rymbr[s+1] = -MAXREAL;
	}
	for ( ; l < n; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rymbr[s] =   mini(rymbr[s],   sml[l].mbr[s]);
		rymbr[s+1] = maxi(rymbr[s+1], sml[l].mbr[s+1]);
	    }
            dead -= area(dimension, sml[l].mbr);
	}
        dead += area(dimension, rymbr);
	over = overlap(dimension, rxmbr, rymbr); 

        if ((over < minover) ||
	    (over == minover) && dead < mindead)
        {
            minover = over;
            mindead = dead;
            dist = m1+k;
            lu = true;
        }

        // upper sort
	// now calculate margin of R1
	// initialize mbr of R1 
        dead = 0.0;
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rxmbr[s] =    MAXREAL;
	    rxmbr[s+1] = -MAXREAL;
	}
	for (l = 0; l < m1+k; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =   mini(rxmbr[s],   smu[l].mbr[s]);
		rxmbr[s+1] = maxi(rxmbr[s+1], smu[l].mbr[s+1]);
	    }
            dead -= area(dimension, smu[l].mbr);
	}
        dead += area(dimension, rxmbr);
	
	// now calculate margin of R2
	// initialize mbr of R2 
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rymbr[s] =    MAXREAL;
	    rymbr[s+1] = -MAXREAL;
	}
	for ( ; l < n; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rymbr[s] =   mini(rymbr[s],   smu[l].mbr[s]);
		rymbr[s+1] = maxi(rymbr[s+1], smu[l].mbr[s+1]);
	    }
            dead -= area(dimension, smu[l].mbr);
	}
        dead += area(dimension, rxmbr);
	over = overlap(dimension, rxmbr, rymbr); 

        if ((over < minover) ||
	    (over == minover) && dead < mindead)
        {
            minover = over;
            mindead = dead;
            dist = m1+k;
            lu = false;
        }
    }

    // calculate best distribution
    *distribution = new int[n];    
    for (i = 0; i < n; i++)
    {
        if (lu)
            (*distribution)[i] = sml[i].index;
        else
            (*distribution)[i] = smu[i].index;
    }

    delete sml;
    delete smu;
    delete rxmbr;
    delete rymbr;


    return dist;
}


//////////////////////////////////////////////////////////////////////////////
// RTDirNode
//////////////////////////////////////////////////////////////////////////////

template <class DATA> 
RTDirNode<DATA>::RTDirNode(RTree<DATA> *rt)
    : RTNode<DATA>(rt)
// zugehoeriger Plattenblock muss erst erzeugt werden
{
    char *b;
    int header_size;
    DirEntry<DATA> * d;

    // mal kurz einen Dateneintrag erzeugen und schauen, wie gross der wird..
    d = new DirEntry<DATA>(this->dimension, rt);

    // von der Blocklaenge geht die Headergroesse ab
    header_size = sizeof(char) + sizeof(char) + sizeof(int);
    this->capacity = (rt->file->get_blocklength() - header_size) / d->get_size();
    delete d;

    // Eintraege erzeugen, das geht mit einem Trick, da C++ beim 
    // Initialisieren von Objektarrays nur Defaultkonstruktoren kapiert.
    // Daher wird ueber globale Variablen die Information uebergeben.
    RTDirNode__dimension = this->dimension;
    RTDirNode__my_tree = this->my_tree;
    entries = new DirEntry<DATA>[this->capacity];

    // neuen Plattenblock an File anhaengen
    b = new char[rt->file->get_blocklength()];
    this->block = rt->file->append_block(b);
    delete b;

    rt->num_of_inodes ++;

    // Plattenblock muss auf jeden Fall neu geschrieben werden
    this->dirty = true;
}


template <class DATA> 
RTDirNode<DATA>::RTDirNode(RTree<DATA> *rt, int _block)
    : RTNode<DATA>(rt)
// zugehoeriger Plattenblock existiert schon
{
    char *b;
    int header_size;
    DirEntry<DATA> * d;

    // mal kurz einen Dateneintrag erzeugen und schauen, wie gross der wird..
    d = new DirEntry<DATA>(this->dimension, rt);

    // von der Blocklaenge geht die Headergroesse ab
    header_size = sizeof(char) + sizeof(char) + sizeof(int);
    this->capacity = (rt->file->get_blocklength() - header_size) / d->get_size();
    delete d;

    // Eintraege erzeugen, das geht mit einem Trick, da C++ beim
    // Initialisieren von Objektarrays nur Defaultkonstruktoren kapiert.
    // Daher wird ueber globale Variablen die Information uebergeben.
    RTDirNode__dimension = this->dimension;
    RTDirNode__my_tree = this->my_tree;
    entries = new DirEntry<DATA>[this->capacity];

    // zugehoerigen Plattenblock holen und Daten einlesen
    // dies kann nicht in RTNode::RTNode(..) passieren, weil
    // beim Aufruf des Basisklassenkonstruktors RTDirNode noch
    // gar nicht konstruiert ist, also auch keine Daten aufnehmen kann
    this->block = _block;
    b = new char[rt->file->get_blocklength()];
    rt->file->read_block(b, this->block);
    read_from_buffer(b);
    delete b;

    // Plattenblock muss vorerst nicht geschrieben werden
    this->dirty = false;
}

template <class DATA> 
RTDirNode<DATA>::~RTDirNode()
{
    char *b;

    if (this->dirty)
    {
	// Daten auf zugehoerigen Plattenblock schreiben
	b = new char[this->my_tree->file->get_blocklength()];
	write_to_buffer(b);
	this->my_tree->file->write_block(b, this->block);
        delete b;
    }

    delete [] entries;
}


template <class DATA> 
void RTDirNode<DATA>::read_from_buffer(char *buffer)
{
    int i, j, s;

    // Lesen, ob Sohnknoten eine Datenseite ist
    memcpy(&son_is_data, buffer, sizeof(bool));
    j = sizeof(bool);

    // Level des Knotens lesen
    memcpy(&this->level, &buffer[j], sizeof(char));
    j += sizeof(char);

    // Anzahl der belegten Eintraege lesen
    memcpy(&this->num_entries, &buffer[j], sizeof(int));
    j += sizeof(int);

    s = entries[0].get_size();
    for (i = 0; i < this->num_entries; i++)
    {
	entries[i].read_from_buffer(&buffer[j]);
	entries[i].son_is_data = son_is_data;
	j += s;
    }
}

template <class DATA> 
void RTDirNode<DATA>::write_to_buffer(char *buffer)
{
    int i, j, s;

    // Schreiben, ob Sohnknoten eine Datenseite ist
    memcpy(buffer, &son_is_data, sizeof(bool));
    j = sizeof(bool);

    // Level des Knotens schreiben
    memcpy(&buffer[j], &this->level, sizeof(char));
    j += sizeof(char);

    // Anzahl der belegten Eintraege schreiben
    memcpy(&buffer[j], &this->num_entries, sizeof(int));
    j += sizeof(int);

    s = entries[0].get_size();
    for (i = 0; i < this->num_entries; i++)
    {
	entries[i].write_to_buffer(&buffer[j]);
	j += s;
    }
}

template <class DATA> 
void RTDirNode<DATA>::print()
{
    int i, n;

    n = this->get_num();
    for (i = 0; i < n ; i++)
    { 
        printf("(%4.1lf, %4.1lf, %4.1lf, %4.1lf)\n", 
	       entries[i].bounces[0],
	       entries[i].bounces[1],
	       entries[i].bounces[2],
	       entries[i].bounces[3]);
    }
    printf("this->level %d\n", this->level);
}

template <class DATA> 
int RTDirNode<DATA>::get_num_of_data()
{
    int i, n, sum;

    n = this->get_num();
    sum = 0;
    for (i = 0; i < n ; i++)
        sum += entries[i].num_of_data;
    return sum;
}


template <class DATA> 
float * RTDirNode<DATA>::get_mbr()
// liefert mbr des Directoryknotens
{
    int i, j, n;
    float *mbr;

    mbr = new float[2*this->dimension];
    for (i = 0; i < 2*this->dimension; i ++ )

	mbr[i] = entries[0].bounces[i];

    n = this->get_num();
    for (j = 1; j < n; j++)
    {
	for (i = 0; i < 2*this->dimension; i += 2)
	{
	    mbr[i]   = mini(mbr[i],   entries[j].bounces[i]);
	    mbr[i+1] = maxi(mbr[i+1], entries[j].bounces[i+1]);
        }
    }

    return mbr;
}

template <class DATA> 
void RTDirNode<DATA>::enter(DirEntry<DATA> *de)
{
    // ist ein Einfuegen ueberhaupt moeglich?
    if (this->get_num() > (this->capacity-1))
        error("RTDirNode::enter: called, but node is full", true);
  
    // Eintrag an erste freie Stelle kopieren
    entries[this->num_entries] = *de;

    // jetzt gibts einen mehr
    this->num_entries++;

    // Eintrag freigeben, aber keine Nachfolgeknoten loeschen !!!!
    de->son_ptr = NULL;
    delete de;
}

template <class DATA> 
void RTDirNode<DATA>::split(RTDirNode<DATA> *sn)
// splittet den aktuellen Knoten so auf, dass m mbr's nach sn verschoben
// werden
{
    int i, *distribution, dist, n;
    float **mbr_array;
    DirEntry<DATA> *new_entries1, *new_entries2;

#ifdef SHOWMBR
	split_000++;
#endif

    // wieviele sind denn nun belegt?
    n = this->get_num();

    // mbr_array allokieren und belegen
    mbr_array = new floatptr[n];
    for (i = 0; i < n; i++)
    { 
       	mbr_array[i] = entries[i].bounces;
    }
    // Verteilung berechnen
    dist = RTNode<DATA>::split(mbr_array, &distribution);

    // mbr_array freigeben
    delete [] mbr_array;

    // neues Datenarray erzeugen
    // --> siehe Konstruktor
    RTDirNode__dimension = this->dimension;
    RTDirNode__my_tree = this->my_tree;
    new_entries1 = new DirEntry<DATA>[this->capacity];
    new_entries2 = new DirEntry<DATA>[this->capacity];

    for (i = 0; i < dist; i++)
    {
       	new_entries1[i] = entries[distribution[i]];
    }
	
    for (i = dist; i < n; i++)
    {
       	new_entries2[i-dist] = entries[distribution[i]];
    }

    // Datenarrays freigeben
    // da die Nachfolgeknoten aber nicht geloescht werden sollen
    // werden vor dem Aufruf von delete noch alle Pointer auf NULL gesetzt
    for (i = 0; i < n; i++)
    {
       	entries[i].son_ptr = NULL;
       	sn->entries[i].son_ptr = NULL;
    }
    delete [] entries;
    delete [] sn->entries;

    // neue Datenarrays kopieren
    entries = new_entries1;
    sn->entries = new_entries2;

    // Anzahl der Eintraege berichtigen
    this->num_entries = dist;
    sn->num_entries = n - dist;  // muss wegen Rundung so bleiben !!
}


template <class DATA> 
int RTDirNode<DATA>::choose_subtree(float *mbr)
{
    int i, j, n, follow, minindex, *inside, inside_count, *over;
    float *bmbr, old_o, o, omin, a, amin, f, fmin;

    n = this->get_num();

    // faellt d in einen bestehenden Eintrag ?
    inside_count = 0;
    inside = new int[n];
    over = new int[n];
    for (i = 0; i < n; i++)
    {
	switch (entries[i].section(mbr))
	{
	case INSIDE:
	    inside[inside_count++] = i;
	    break;
	}
    }

    if (inside_count == 1)
    // Fall 1: Rechteck faellt genau in einen Eintrag
	follow = inside[0];
    else if (inside_count > 1)
    // Fall 2: Rechteck faellt in mehrere Eintrag --> nimm den
    // mit der geringsten Flaeche (volumen!!!)
    {
	fmin = MAXFLOAT;
	//printf("Punkt in %d von %d MBRs \n",inside_count,n);
	for (i = 0; i < inside_count; i++)
	{
	    f = area(this->dimension, entries[inside[i]].bounces);
	    if (f < fmin)
	    {
		minindex = i;
      		fmin = f;
       	    }
       	}
	follow = inside[minindex];
    }
    else
    // Fall 3: Rechteck faellt in keinen Eintrag -->
    // fuer Knoten, die auf interne Knoten zeigen:
    // nimm den Eintrag, der am geringsten vergroessert wird;
    // bei gleicher Vergroesserung:
    // nimm den Eintrag, der die geringste Flaeche hat
    //
    // fuer Knoten, die auf Datenknoten zeigen:
    // nimm den, der die geringste Ueberlappung verursacht
    // bei gleicher Ueberlappung:
    // nimm den Eintrag, der am geringsten vergroessert wird;
    // bei gleicher Vergroesserung:
    // nimm den Eintrag, der die geringste Flaeche hat
    {
       	if (son_is_data)
	{
            omin = MAXREAL;
	    fmin = MAXREAL;
	    amin = MAXREAL;
	    for (i = 0; i < n; i++)
	    {
		// berechne die mbr, wenn mbr in entries[i] eingefuegt wird
		enlarge(this->dimension, &bmbr, mbr, entries[i].bounces);

		// calculate area and area enlargement
		a = area(this->dimension, entries[i].bounces);
		f = area(this->dimension, bmbr) - a;

		// calculate overlap before enlarging entry_i
		old_o = o = 0.0;
		for (j = 0; j < n; j++)
		{
		    if (j != i)
		    {
			old_o += overlap(this->dimension,
					 entries[i].bounces,
					 entries[j].bounces);
			o += overlap(this->dimension,
				     bmbr,
				     entries[j].bounces);
		    }
	        }
	        o -= old_o;

	        // is this entry better than the former optimum ?
	        if ((o < omin) ||
		    (o == omin && f < fmin) ||
		    (o == omin && f == fmin && a < amin))
	        {
	       	    minindex = i;
		    omin = o;
		    fmin = f;
		    amin = a;
	        }
	        delete bmbr;
	    }
        }
        else
        {
	    fmin = MAXREAL;
	    amin = MAXREAL;
	    for (i = 0; i < n; i++)
	    {
	        // berechne die mbr, wenn mbr in entries[i] eingefuegt wird
	        enlarge(this->dimension, &bmbr, mbr, entries[i].bounces);

	        // calculate area and area enlargement
	        a = area(this->dimension, entries[i].bounces);
	        f = area(this->dimension, bmbr) - a;

	        // is this entry better than the former optimum ?
	        if ((f < fmin) ||
		    (f == fmin && a < amin))
	        {
	       	    minindex = i;
		    fmin = f;
	            amin = a;
	        }
	        delete bmbr;
	    }
        }

	// berechne die neue mbr und schreib sie in den Eintrag
	enlarge(this->dimension, &bmbr, mbr, entries[minindex].bounces);
	memcpy(entries[minindex].bounces, bmbr,
	       this->dimension * 2 * sizeof(float));
	follow = minindex;
	delete bmbr;

	// leider muss jetzt der Block auch geschrieben werden
	this->dirty = true;
    }

    delete inside;
    delete over;

    return follow;
}

template <class DATA> 
R_OVERFLOW RTDirNode<DATA>::insert(DATA *d, RTNode<DATA> **sn)
{
    int follow;
    RTNode<DATA> *succ, *new_succ;
    RTDirNode<DATA> *brother;
    DirEntry<DATA> *de;
    R_OVERFLOW ret;
    float *mbr,*nmbr;

    // Einfuegepfad waehlen
    mbr = d->get_mbr();
    follow = choose_subtree(mbr);
    delete[] mbr;

    // Sohn laden
    succ = entries[follow].get_son();

    // insert d into son
    ret = succ->insert(d, &new_succ);
    if (ret != NONE)
    // if anything happend --> update bounces of entry "follow"
    {
        mbr = succ->get_mbr();
        memcpy(entries[follow].bounces, mbr, sizeof(float) * 2*this->dimension);
        delete[] mbr;
    }

    // recalculate # of succeeders in the tree
    entries[follow].num_of_data = succ->get_num_of_data();

    if (ret == SPLIT)
    // Einfuegen von d in Sohn hat zum Split des Sohns gefuehrt -->
    // neuen Sohn *new_succ einfuegen
    {
        if (this->get_num() == this->capacity)
     	    error("RTDataNode::insert: maximum this->capacity violation", true);

         // create and insert new entry
        de = new DirEntry<DATA>(this->dimension, this->my_tree);
	nmbr = new_succ->get_mbr();
        memcpy(de->bounces, nmbr, 2*this->dimension*sizeof(float));
	delete nmbr;
        de->son = new_succ->block;
        de->son_ptr = new_succ;
        de->son_is_data = son_is_data;
        de->num_of_data = new_succ->get_num_of_data();
        enter(de);

        if (this->get_num() == (this->capacity - 1))
        // Knoten laeuft ueber --> Split
        // this happens already if the node is nearly filled
        // for the algorithms are more easy then
        {
	    brother = new RTDirNode<DATA>(this->my_tree);
	    brother->son_is_data = son_is_data;
	    brother->level = this->level;
	    split(brother);
            *sn = brother;
/*	    printf("dddiiirrr: splitting node %d, creating %d\n",
		   this->block, (*sn)->block);
*/
		    
            ret = SPLIT;
	}
        else
      	    ret = NONE;
    }
    // must write page
    this->dirty = true;

    return ret;
}

template <class DATA> 
DATA * RTDirNode<DATA>::get(int i, int dcopy)
{
    int j, n, sum;
    RTNode<DATA> *sn;

    n = this->get_num();
    sum = 0;
    for (j = 0; j < n; j++)
    {
        sum += entries[j].num_of_data;

        if (sum > i)
	// i-th object is behind this node --> follow son
        {
            sn = entries[j].get_son();
            return sn->get(i - (sum - entries[j].num_of_data), dcopy);
        }
    }

    return NULL;
}

template <class DATA> 
void RTDirNode<DATA>::region(float *mbr)
{
    int i, n;
    SECTION s;
    RTNode<DATA> *succ;

    n = this->get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        s = entries[i].section(mbr);
        if (s == INSIDE || s == OVERLAP)
        {
            // Rechteck ist interessant --> rekursiv weiter
            succ = entries[i].get_son();
            succ->region(mbr);
        }
    }
}


template <class DATA> 
void RTDirNode<DATA>::rangeQuery(float *mbr, SortedLinList<DATA> *res)
{
    int i, n;
    SECTION s;
    RTNode<DATA> *succ;

#ifdef ZAEHLER
    page_access += this->my_tree->node_weight[this->level];
#endif

    n = this->get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        s = entries[i].section(mbr);
        if (s == INSIDE || s == OVERLAP)
        {
            // Rechteck ist interessant --> rekursiv weiter
            succ = entries[i].get_son();
            succ->rangeQuery(mbr,res);
        }
    }
}

template <class DATA> 
void RTDirNode<DATA>::rangeQuery(float *mbr, LinList<DATA> *res)
{
    int i, n;
    SECTION s;
    RTNode<DATA> *succ;

#ifdef ZAEHLER
    page_access += this->my_tree->node_weight[this->level];
#endif

    n = this->get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        s = entries[i].section(mbr);
        if (s == INSIDE || s == OVERLAP)
        {
            // Rechteck ist interessant --> rekursiv weiter
            succ = entries[i].get_son();
            succ->rangeQuery(mbr,res);
        }
    }
}

template <class DATA> 
void RTDirNode<DATA>::rangeQuery(DATA *center, float radius, 
		SortedLinList<DATA> *res)
{
    int i, n;
    bool s;
    RTNode<DATA> *succ;

    #ifdef ZAEHLER
	page_access += this->my_tree->node_weight[this->level];
    #endif

    n = this->get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        s = entries[i].section_circle(center,radius);
        if (s)
        {
            // Rechteck ist interessant --> rekursiv weiter
            succ = entries[i].get_son();
            succ->rangeQuery(center,radius,res);
        }
    }
}

template <class DATA>
void RTDirNode<DATA>::NearestNeighborSearch(DATA *QueryPoint, DATA *&Nearest,
				float &nearest_distanz)
{
    float *minmax_distanz;		// Array fuer MINMAXDIST aller Eintr"age
    int *indexliste;		// Liste (for Sorting and Prunching)
    int i,j,k,last,n;
    float akt_min_dist;		// minimal distanz computed upto now
    float minmaxdist,mindist;
    BranchList *activebranchList;
#ifdef ZAEHLER
    page_access += this->my_tree->node_weight[this->level];
#endif
    n = this->get_num();
    activebranchList = new BranchList [n]; // Array erzeugen mit n Elementen
    for( i = 0; i < n; i++)
    {
	activebranchList[i].entry_number = i;
	activebranchList[i].minmaxdist = MINMAXDIST(QueryPoint,entries[i].bounces);
	activebranchList[i].mindist = MINDIST(QueryPoint,entries[i].bounces);
    }
    // sort branchList
    qsort(activebranchList,n,sizeof(BranchList),sortmindist);
    // prune BrunchList
    last = pruneBrunchList(nearest_distanz,activebranchList,n);
    for( i = 0; i < last; i++)
    {
	entries[activebranchList[i].entry_number].get_son()->NearestNeighborSearch(QueryPoint, Nearest, nearest_distanz);
 	
	last = pruneBrunchList(nearest_distanz,activebranchList,last,false);
    }
    delete activebranchList;
}

template <class DATA>
void RTDirNode<DATA>::NearestNeighborSearch(DATA *QueryPoint,
				SortedLinList<DATA> *&res,
				float &nearest_distanz)
{
    float *minmax_distanz;		// Array fuer MINMAXDIST aller Eintr"age
    int *indexliste;		// Liste (for Sorting and Prunching)
    int i,j,k,last,n;
    float akt_min_dist;		// minimal distanz computed upto now 
    float minmaxdist,mindist;
    
    BranchList *activebranchList;
    
#ifdef ZAEHLER
    page_access += this->my_tree->node_weight[this->level];
#endif

    n = this->get_num();
    
    k = res->get_num(); 	// wird haben eine k-nearest-Narbor-Query
    
    nearest_distanz = res->get(k-1)->distanz;	// der aktuell letzte 
                                                // n"achste Nachbar wird
                                                // versucht zu ersetzen.

    activebranchList = new BranchList [n]; // Array erzeugen mit n Elementen
 
    for( i = 0; i < n; i++)
    {
	activebranchList[i].entry_number = i;
	activebranchList[i].minmaxdist = MINMAXDIST(QueryPoint,entries[i].bounces);
	activebranchList[i].mindist = MINDIST(QueryPoint,entries[i].bounces);	
    }

    // sortbranchList
    qsort(activebranchList,n,sizeof(BranchList),sortmindist);

    // pruneBrunchList
    last = this->kNN_pruneBrunchList(nearest_distanz,activebranchList,res,n);

    for( i = 0; i < last; i++)
    {
	entries[activebranchList[i].entry_number].get_son()->NearestNeighborSearch(QueryPoint, res, nearest_distanz);

	last = this->kNN_pruneBrunchList(nearest_distanz,activebranchList,res,last,false);
    }

    delete activebranchList;
}

template <class DATA>
int RTDirNode<DATA>::kNN_pruneBrunchList(float &nearest_distanz,
			const void *activebrunchList,
			SortedLinList<DATA> *res,
			int n, int downward)
{
    // Schneidet im Array BranchList alle Eintraege ab, deren Distanz groesser
    // ist als die aktuell naeheste Distanz
    //

    BranchList *bl; 
    int i, j, aktlast;
    bl = (BranchList *) activebrunchList;

if ( downward )
 {
    aktlast=n;

    // 2. Strategie:
    //
    // nearest_distanz > MINMAXDIST(P,M)
    // -> nearest_distanz = MIMMAXDIST(P,M)
    // This is also ONLY used in downward pruning!!!

    SortedLinList<DATA> cres;
    DATA *d1, *d2, *letzte;
    bool t;
    int k;

    cres.set_sorting(true);

    for ( d1 = res->get_first(); d1; d1 = res->get_next()) {
        d2 = new DATA(&this->dimension);
        *d2 = *d1;
        cres.insert(d2);
     }

    k = cres.get_num();

    for( i = 0; i < aktlast; i++)
	if (nearest_distanz > bl[i].minmaxdist) {
	    // l"osche letztes Elemente der res-Liste
	    letzte = cres.get(k-1);
	    t = cres.erase();
	    letzte = new DATA(&this->dimension);

	    // Erzeuge neuen Punkt in der res-Liste
	    for(int l = 0; l < this->dimension; l++)
		letzte->data[l] = MAXREAL;
	    letzte->distanz = bl[i].minmaxdist;
	    cres.insert(letzte);

	    nearest_distanz = cres.get(k-1)->distanz;
	 }
 }
 else
  {
    // 3. Strategie:
    // 
    // nearest_distanz < MINDIST(P,M) <= ||(P,o)||
    //
    // in M kann der Nearest-Narbor sicher nicht mehr liegen.
    // This is ONLY used in upward pruning!!!

    aktlast = n;


    for( i = 0; (i < aktlast) && (nearest_distanz >= bl[i].mindist) ; i++)
     {
     };
    aktlast = i;

  }
    return (aktlast);
}

template <class DATA>
void RTDirNode<DATA>::range_nnQuery(float *mbr, SortedLinList<DATA> *res,
				DATA *center, float &nearest_distanz,
				DATA *Nearest, bool *success)
{
    float *minmax_distanz;		// Array fuer MINMAXDIST aller Eintr"age
    int *indexliste;		// Liste (for Sorting and Prunching)
    int i,j,k,last,n;
    float akt_min_dist;		// minimal distanz computed upto now
    float minmaxdist,mindist;
    SECTION s;
    
    BranchList *activebranchList, *sectionList;
    
    if (*success)
    {
	rangeQuery(mbr,res);
 	return;
    }
    
#ifdef ZAEHLER
    page_access += this->my_tree->node_weight[this->level];
#endif



    n = this->get_num();

    sectionList = new BranchList [n];
    
    // Erst einmal feststellen, welche der Eintr"age einen Schnitt mit dem 
    // aktuellen mbr haben und Ergebnis in sectionList speichern
    
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        s = entries[i].section(mbr);
	sectionList[i].entry_number = i;
        if (s == INSIDE || s == OVERLAP)
	    sectionList[i].section  = true;            
	else
	    sectionList[i].section  = false;   
    }


    activebranchList = new BranchList [n]; // Array erzeugen mit n Elementen
 
    for( i = 0; i < n; i++)
    {
	activebranchList[i].entry_number = i;
	activebranchList[i].minmaxdist = MINMAXDIST(center,entries[i].bounces);
	activebranchList[i].mindist = MINDIST(center,entries[i].bounces);	
    }	
    
    // sortbranchList
    qsort(activebranchList,n,sizeof(BranchList),sortmindist); 
    
    // pruneBrunchList
    last = pruneBrunchList(nearest_distanz,activebranchList,n);

    // jetzt nachsehen, ob auch Schnittmenge mit mbr abgeschnitten wurde:
    // falls ja: Schnitt wieder R"ueckg"angig machen
  
    last = testBrunchList(activebranchList,sectionList,n,last);

    for( i = 0; i < last; i++)
    {
	entries[activebranchList[i].entry_number].get_son()->
	    range_nnQuery(mbr,res,center,nearest_distanz,
			  Nearest,success);
 	
	last = pruneBrunchList(nearest_distanz,activebranchList,last,false);
        last = testBrunchList(activebranchList,sectionList,n,last,false);
    }
    
    delete [] activebranchList;
    delete [] sectionList;
}

template <class DATA>
void RTDirNode<DATA>::overlapping(float *p, int *nodes_t)
{
    int i, n;
    RTNode<DATA> *succ;

    // ein Knoten mehr besucht
    nodes_t[this->level]++;

    n = this->get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        if (entries[i].is_inside(p))
        {
            // Rechteck ist interessant --> rekursiv weiter
            succ = entries[i].get_son();
            succ->overlapping(p, nodes_t);
        }
    }
}

template <class DATA>
void RTDirNode<DATA>::nodes(int *nodes_a)
{
    int i, n;
    RTNode<DATA> *succ;

    // ein Knoten mehr besucht
    nodes_a[this->level]++;

    n = this->get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
	succ = entries[i].get_son();
	succ->nodes(nodes_a);
    }
}

template <class DATA> 
void RTDirNode<DATA>::writeinfo(FILE *f)
{
    int i,j,n;

    float *mbr;

    mbr = get_mbr();
    fprintf(f,"%d\n",this->level+1);
    fprintf(f,"move %f %f\n",mbr[0],mbr[2]);
    fprintf(f,"draw %f %f\n",mbr[1],mbr[2]);
    fprintf(f,"draw %f %f\n",mbr[1],mbr[3]);
    fprintf(f,"draw %f %f\n",mbr[0],mbr[3]);
    fprintf(f,"draw %f %f\n",mbr[0],mbr[2]);
    fprintf(f,"\n");
    delete[] mbr;

    n = this->get_num();
    for( i = 0; i < n ; i++)
	entries[i].get_son()->writeinfo(f);

}

template <class DATA>
void RTDirNode<DATA>::get_dmbr(float* dmbr,int& length)
{
    int n = this->get_num();
    for ( int i = 0; i < n; i++ )
	entries[i].get_son()->get_dmbr(dmbr,length);
}

//////////////////////////////////////////////////////////////////////////////
// RTDataNode
//////////////////////////////////////////////////////////////////////////////

template <class DATA> 
RTDataNode<DATA>::RTDataNode(RTree<DATA> *rt)
    : RTNode<DATA>(rt)
// zugehoeriger Plattenblock muss erst erzeugt werden
{
    char *b;
    int header_size;
    DATA * d;

    this->level = 0;

    // mal kurz einen Dateneintrag erzeugen und schauen, wie gross der wird..
    d = new DATA(&this->dimension);

    // von der Blocklaenge geht die Headergroesse ab
    header_size = sizeof(char) + sizeof(int);
    this->capacity = (rt->file->get_blocklength() - header_size) / d->get_size();

    delete d;

    // Eintraege erzeugen; das geht mit einem Trick, da C++ beim 
    // Initialisieren von Objektarrays nur Defaultkonstruktoren kapiert.
    // Daher wird ueber statische Variablen die Information uebergeben.
    RTDataNode__dimension = this->dimension;
    data = new DATA[this->capacity];

    // neuen Plattenblock an File anhaengen
    b = new char[rt->file->get_blocklength()];
    this->block = rt->file->append_block(b);
    delete b;

    rt->num_of_dnodes ++;

    // Plattenblock muss auf jeden Fall neu geschrieben werden
    this->dirty = true;
}

template <class DATA> 
RTDataNode<DATA>::RTDataNode(RTree<DATA> *rt, int _block)
    : RTNode<DATA>(rt)
// zugehoeriger Plattenblock existiert schon
{
    char *b;
    int header_size;
    DATA *d;

    // mal kurz einen Dateneintrag erzeugen und schauen, wie gross der wird..
    d = new DATA(&this->dimension);

    // von der Blocklaenge geht die Headergroesse ab
    header_size = sizeof(char) + sizeof(int);
    this->capacity = (rt->file->get_blocklength() - header_size) / d->get_size();

    delete d;

    // Eintraege erzeugen; das geht mit einem Trick, da C++ beim 
    // Initialisieren von Objektarrays nur Defaultkonstruktoren kapiert.
    // Daher wird ueber statische Variablen die Information uebergeben.
    RTDataNode__dimension = this->dimension;
    data = new DATA[this->capacity];

    // zugehoerigen Plattenblock holen und Daten einlesen
    // dies kann nicht in RTNode::RTNode(..) passieren, weil
    // beim Aufruf des Basisklassenkonstruktors RTDirNode noch
    // gar nicht konstruiert ist, also auch keine Daten aufnehmen kann
    this->block = _block;
    b = new char[rt->file->get_blocklength()];
    rt->file->read_block(b, this->block);
    read_from_buffer(b);
    delete b;

    // Plattenblock muss vorerst nicht geschrieben werden
    this->dirty = false;
}

template <class DATA> 
RTDataNode<DATA>::~RTDataNode()
{
    char *b;

    if (this->dirty)
    {
	// Daten auf zugehoerigen Plattenblock schreiben
	b = new char[this->my_tree->file->get_blocklength()];
	write_to_buffer(b);
	this->my_tree->file->write_block(b, this->block);
        delete b;
    }

    delete [] data;
}

template <class DATA> 
void RTDataNode<DATA>::read_from_buffer(char *buffer)
{
    int i, j, s;

    // Level des Knotens lesen
    memcpy(&this->level, buffer, sizeof(char));
    j = sizeof(char);

///////////////////////////////////////////////////////////
    // da das Abspeichern des levels voellig ueberfluessig ist
    // und this->level manchmal falsch in den Dateien steht, dieser HACK:
    this->level = 0;
///////////////////////////////////////////////////////////

    // Anzahl der belegten Eintraege lesen
    memcpy(&this->num_entries, &buffer[j], sizeof(int));
    j += sizeof(int);

    s = data[0].get_size();
    for (i = 0; i < this->num_entries; i++)
    {
	data[i].read_from_buffer(&buffer[j]);
	j += s;
    }
}

template <class DATA> 
void RTDataNode<DATA>::write_to_buffer(char *buffer)
{
    int i, j, s;

    // Level des Knotens schreiben
    this->level = 0;
    memcpy(&this->level, buffer, sizeof(char));
    j = sizeof(char);

    // Anzahl der belegten Eintraege schreiben
    memcpy(&buffer[j], &this->num_entries, sizeof(int));
    j += sizeof(int);

    s = data[0].get_size();
    for (i = 0; i < this->num_entries; i++)
    {
	data[i].write_to_buffer(&buffer[j]);
	j += s;
    }
}

template <class DATA> 
void RTDataNode<DATA>::print()
{
    int i, n;

    n = this->get_num();
    for (i = 0; i < n ; i++)
    { 
        printf("(%4.1lf, %4.1lf)\n", 
	       data[i].data[0],
	       data[i].data[1]);
    }
    printf("this->level %d\n", this->level);
}


template <class DATA> 
int RTDataNode<DATA>::get_num_of_data()
{
    return this->get_num();
}

template <class DATA> 
float * RTDataNode<DATA>::get_mbr()
// liefert mbr des Datenknotens
{
    int i, j, n;
    float *mbr, *tm;

    mbr = data[0].get_mbr();
    n = this->get_num();
    for (j = 1; j < n; j++)
    {
	tm = data[j].get_mbr();
	for (i = 0; i < 2*this->dimension; i += 2)
	{
		mbr[i]   = mini(mbr[i],   tm[i]);
		mbr[i+1] = maxi(mbr[i+1], tm[i+1]);
       	}
       	delete tm;
    }

    return mbr;
}


template <class DATA> 
void RTDataNode<DATA>::split(RTDataNode<DATA> *sn)
// splittet den aktuellen Knoten so auf, dass m mbr's nach sn verschoben
// werden
{
    int i, *distribution, dist, n;
    float **mbr_array;
    DATA *new_data1, *new_data2;

#ifdef SHOWMBR
	split_000++;
#endif

    // wieviele sind denn nun belegt?
    n = this->get_num();

    // mbr_array allokieren und belegen
    mbr_array = new floatptr[n];
    for (i = 0; i < n; i++)
    {
        mbr_array[i] = data[i].get_mbr();
    }
    
    // Verteilung berechnen 
    dist = RTNode<DATA>::split(mbr_array, &distribution);

    // neues Datenarray erzeugen
    // --> siehe Konstruktor
    RTDataNode__dimension = this->dimension;
    new_data1 = new DATA[this->capacity];
    new_data2 = new DATA[this->capacity];

    for (i = 0; i < dist; i++)
        new_data1[i] = data[distribution[i]];

    for (i = dist; i < n; i++)
        new_data2[i-dist] = data[distribution[i]];

    // Datenarrays freigeben
    delete [] data;
	delete [] ((RTDataNode *)sn)->data;

    // neue Datenarrays kopieren
    data = new_data1;
    sn->data = new_data2;

    // Anzahl der Eintraege berichtigen
    this->num_entries = dist;
    sn->num_entries = n - dist;  // muss wegen Rundung so bleiben !!

    // mbr_array freigeben
    for (i = 0; i < n; i++)
	delete [] mbr_array[i];

    delete [] mbr_array;
}


template <class DATA> 
R_OVERFLOW RTDataNode<DATA>::insert(DATA *d, RTNode<DATA> **sn)
// liefert false, wenn die Seite gesplittet werden muss
{
    int i, last_cand;
    float *mbr, *center;
    SortMbr *sm;
    DATA *nd, *new_data;

    if (this->get_num() == this->capacity)
	error("RTDataNode::insert: maximum this->capacity violation", true);

    data[this->get_num()] = *d;
    this->num_entries ++;

    // Plattenblock zum Schreiben markieren
    this->dirty = true;

    if (this->get_num() == (this->capacity - 1))
    // Knoten laeuft ueber --> Split
    // dies passiert schon beim vorletzten Datum, da dadurch 
    // die Algorithmen einfacher werden
    {
        if (this->my_tree->re_level[0] == false)
	// there was no reinsert on this->level 0 during this insertion
        {
            // calculate center of page
            mbr = get_mbr();
            center = new float[this->dimension];
            for (i = 0; i < this->dimension; i++)
                 center[i] = (mbr[2*i] + mbr[2*i+1]) / 2.0;

            // neues Datenarray erzeugen
            // --> siehe Konstruktor
            RTDataNode__dimension = this->dimension;
            new_data = new DATA[this->capacity];

	    // Indexarray aufbauen und initialisieren
	    sm = new SortMbr[this->num_entries];    
	    for (i = 0; i < this->num_entries; i++)
	    {
		sm[i].index = i;
		sm[i].dimension = this->dimension;
		sm[i].mbr = data[i].get_mbr();
		sm[i].center = center;
            }

            // sort by distance of each center to the overall center
            qsort(sm, this->num_entries, sizeof(SortMbr), sort_center_mbr);

            last_cand = (int) ((float)this->num_entries * 0.30);

            // copy the nearest candidates to new array
            for (i = 0; i < this->num_entries - last_cand; i++)
	        new_data[i] = data[sm[i].index];

            // insert candidates into reinsertion list
            for ( ; i < this->num_entries; i++)
            {
                nd = new DATA(&this->dimension);
                *nd = data[sm[i].index];
                this->my_tree->re_data_cands.insert(nd);
            }

            // free and copy data array
            delete [] data;
	    data = new_data;

       	    for (i = 0; i < this->num_entries; i++)
	       	delete sm[i].mbr;
	    delete sm;
	    delete[] mbr;
	    delete center;
	    this->my_tree->re_level[0] = true;

	    // correct # of entries
	    this->num_entries -= last_cand;

	    // must write page
	    this->dirty = true;

            return REINSERT;
	}
       	else
       	{
	    *sn = new RTDataNode<DATA>(this->my_tree);
	    (*sn)->level = this->level;
	    split((RTDataNode *) *sn);
//	    printf("splitting node %d, creating %d\n", this->block, (*sn)->block);
	}
	return SPLIT;
    }
    else
        return NONE;
}

template <class DATA> 
DATA * RTDataNode<DATA>::get(int i, int dcopy)
{
    DATA *d;

    if (i >= this->get_num())
    // if there is no i-th object --> NULL
        return NULL;

    if (dcopy) {
      d = new DATA(&this->dimension);
      *d = data[i];		// give the copy of data[i] as result;
    }
    else {
      d = &data[i];
      this->dirty = true;		// give the original data[i] as result;
    }
    return d;
}

template <class DATA> 
void RTDataNode<DATA>::region(float *mbr)
{
    int i, n,j;
    float *dmbr;

    n = this->get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        dmbr = data[i].get_mbr();
        if (section(this->dimension, dmbr, mbr))
	{
/*		printf("( ");
		for( j = 0; j < this->dimension; j++)
			printf(" %f",dmbr[j]);

		printf(" ) \n");
*/	}
        delete[] dmbr;
    }
}

template <class DATA> 
void RTDataNode<DATA>::rangeQuery(float *mbr, SortedLinList<DATA> *res)
{
    int i, n,j;
    float *dmbr;
    DATA *point, *center;

#ifdef ZAEHLER
    page_access += this->my_tree->node_weight[0];
#endif

    n = this->get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        dmbr = data[i].get_mbr();
        if (section(this->dimension, dmbr, mbr))
	{
		point = new DATA(&this->dimension);
                *point = data[i];
		// Center des MBRs ausrechnen         
		center = new DATA(&this->dimension);        
		for(j = 0; j < this->dimension; j++)        
		    center->data[j] = (mbr[2*j] + mbr[2*j+1]) / 2;
		point->distanz = objectDIST(point,center);
		res->insert(point);
		delete center;
	}	
        delete [] dmbr;
    }
}

template <class DATA> 
void RTDataNode<DATA>::rangeQuery(float *mbr, LinList<DATA> *res)
{
    int i, n,j;
    float *dmbr;
    DATA *point, *center;

#ifdef ZAEHLER
    page_access += this->my_tree->node_weight[0];
#endif

    n = this->get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        dmbr = data[i].get_mbr();
        if (section(this->dimension, dmbr, mbr))
	{
		point = new DATA(&this->dimension);
                *point = data[i];
		res->insert(point);
	}	
        delete [] dmbr;
    }
}

template <class DATA> 
void RTDataNode<DATA>::rangeQuery(DATA *center, float radius,
			 SortedLinList<DATA> *res)
{
    int i, n,j,k;
    float *dmbr;
    DATA *point;

#ifdef ZAEHLER
    page_access += this->my_tree->node_weight[0];
#endif

    n = this->get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        dmbr = data[i].get_mbr();
        if (section_c(this->dimension, dmbr, center,radius))
	{
		point = new DATA(&this->dimension);
                *point = data[i];
		point->distanz = objectDIST(point,center);
		res->insert(point);		        	
	}	
        delete [] dmbr;
    }
}

template <class DATA>
void RTDataNode<DATA>::NearestNeighborSearch(DATA *QueryPoint, DATA *&Nearest,
			float &nearest_distanz)
{
    int i,j;
    float nearest_dist,distanz;
    
    //nearest_dist = objectDIST(QueryPoint,Nearest);

#ifdef ZAEHLER
    page_access += this->my_tree->node_weight[0];
#endif

    for(i = 0; i < this->get_num(); i++)
    {
	distanz = objectDIST(QueryPoint,get(i));
	
	if (distanz <= nearest_distanz)
	{
	    nearest_distanz = distanz;
	    Nearest = get(i);
	    Nearest->distanz = distanz;  
	}
    }
}

template <class DATA>
void RTDataNode<DATA>::NearestNeighborSearch(DATA *QueryPoint, 
			SortedLinList<DATA> *&res,
			float &nearest_distanz)
{
    int i,j,k;
    float nearest_dist,distanz;
    bool t;
    DATA *neu,*dummy,*element;
    
#ifdef ZAEHLER
    page_access += this->my_tree->node_weight[0];
#endif
    
    k = res->get_num();
    
    for (i = 0; i < this->get_num(); i++)
    {
	element = get(i);
	distanz = objectDIST(QueryPoint,element);
	
	if (distanz <= nearest_distanz)
	{
	    // l"osche letztes Elemente der res-Liste		
	    dummy = res->get(k-1);
	    t = res->erase();
	    
	    // Erzeuge neuen Punkt in der res-Liste
	    element->distanz = distanz;
	    res->insert(element);
	    
	    nearest_distanz = res->get(k-1)->distanz;			
	}
	else
	  delete element;
    }
}

template <class DATA>
void RTDataNode<DATA>::range_nnQuery(float *mbr,SortedLinList<DATA> *res,
				DATA *center, float &nearest_distanz,
				DATA *Nearest, bool *success)
			
{
    int i,j,n;
    float nearest_dist,distanz,*dmbr;
    DATA *point;
    
    if (*success)
    {
	rangeQuery(mbr,res);
	return;
    }
    
#ifdef ZAEHLER
    page_access += this->my_tree->node_weight[0];
#endif

    n = this->get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        dmbr = data[i].get_mbr();
        if (section(this->dimension, dmbr, mbr))
	{
	    point = new DATA(&this->dimension);
	    *point = data[i];
	    point->distanz = objectDIST(point,center);
	    res->insert(point);
	    *success = true;		        	
	}	
        delete [] dmbr;
    } 

    if (!*success)
    {
	for(i = 0; i < this->get_num(); i++)
	{
	    distanz = objectDIST(center,get(i));
	    
	    if (distanz <= nearest_distanz)
	    {
		nearest_distanz = distanz;
		for(j = 0; j < this->dimension; j++)
		    Nearest->data[j] = get(i)->data[j];
	    }
	}
    }   	
}



template <class DATA> 
void RTDataNode<DATA>::writeinfo(FILE *f)
{
    float *mbr;

    mbr = get_mbr();
    fprintf(f,"%d\n",this->level+1);
    fprintf(f,"move %f %f\n",mbr[0],mbr[2]);
    fprintf(f,"draw %f %f\n",mbr[1],mbr[2]);
    fprintf(f,"draw %f %f\n",mbr[1],mbr[3]);
    fprintf(f,"draw %f %f\n",mbr[0],mbr[3]);
    fprintf(f,"draw %f %f\n",mbr[0],mbr[2]);
    fprintf(f,"\n");
    delete[] mbr;
}

template <class DATA>
void RTDataNode<DATA>::get_dmbr(float* dmbr,int& length)
{
    float *mbr;
    mbr = get_mbr(); 
    dmbr[length]   = mbr[0];	// x0: x_min;
    dmbr[length+1] = mbr[1];	// x1: x_max;
    dmbr[length+2] = mbr[2];	// y0: y_min;
    dmbr[length+3] = mbr[3];	// y1: y_max;
    length = length + 4;
	delete[] mbr;
}

//////////////////////////////////////////////////////////////////////////////
// RTree
//////////////////////////////////////////////////////////////////////////////

template <class DATA> 
RTree<DATA>::RTree(char *fname, int _b_length, int cache_size, int _dimension)
// neuen R-Baum konstruieren
{
	 prototype=false;
    file = new CachedBlockFile(fname, _b_length, cache_size);

    // header allokieren und lesen
    header = new char [file->get_blocklength()];    

    // das ist zwar an sich hier Quatsch, aber damit wird die Variable
    // user_header richtig gesetzt, so dass beim Schreiben des Headers der
    // Aufrufer kein problem hat
    read_header(header);

    dimension = _dimension;
    root = 0;
    root_ptr = NULL;
    root_is_data = true;
    num_of_data = num_of_inodes = num_of_dnodes = 0;

    root_ptr = new RTDataNode<DATA>(this);
    root = root_ptr->block;
}

template <class DATA> 
RTree<DATA>::RTree(char *fname, int cache_size)
// bereits vorhandenen R-Baum laden
{
    int j;

	 prototype=false;    
    // file existiert schon --> Blockgroesse wird im Konstruktor ignoriert
    file = new CachedBlockFile(fname, 0, cache_size);

    // header allokieren und lesen
    header = new char [file->get_blocklength()];    
    file->read_header(header);

    read_header(header);

    root_ptr = NULL;

    if (get_num() == 0)
    // Baum war leer -> Datenknoten anlegen und d einfuegen 
    {
	root_ptr = new RTDataNode<DATA>(this);
	root = root_ptr->block;
	root_ptr->level = 0;
    }
    else 
	load_root();
}

template <class DATA> 
RTree<DATA>::~RTree()
{
	if (prototype) return;
   
    int j;

    write_header(header);

    file->set_header(header);
    delete header;

    if (root_ptr != NULL)
    {
        delete root_ptr;
        root_ptr = NULL;
    }
    delete file;

    /*printf("saved R-Tree containing %d internal, %d data nodes and %d data\n",
	   num_of_inodes, num_of_dnodes, num_of_data);*/
}

template <class DATA>
void RTree<DATA>::printinfo(FILE *f)
{
    fprintf(f,"\n internal: %d data-nodes : %d Datas: %d \n", 
	num_of_inodes, num_of_dnodes, num_of_data);
}

template <class DATA>
void RTree<DATA>::read_header(char *buffer)
{
    int i;

    memcpy(&dimension, buffer, sizeof(dimension));
    i = sizeof(dimension);

    memcpy(&num_of_data, &buffer[i], sizeof(num_of_data));
    i += sizeof(num_of_data);

    memcpy(&num_of_dnodes, &buffer[i], sizeof(num_of_dnodes));
    i += sizeof(num_of_dnodes);

    memcpy(&num_of_inodes, &buffer[i], sizeof(num_of_inodes));
    i += sizeof(num_of_inodes);

    memcpy(&root_is_data, &buffer[i], sizeof(root_is_data));
    i += sizeof(root_is_data);

    memcpy(&root, &buffer[i], sizeof(root));
    i += sizeof(root);

    user_header = &buffer[i];
}


template <class DATA>
void RTree<DATA>::write_header(char *buffer)
{
    int i;

    memcpy(buffer, &dimension, sizeof(dimension));
    i = sizeof(dimension);

    memcpy(&buffer[i], &num_of_data, sizeof(num_of_data));
    i += sizeof(num_of_data);

    memcpy(&buffer[i], &num_of_dnodes, sizeof(num_of_dnodes));
    i += sizeof(num_of_dnodes);

    memcpy(&buffer[i], &num_of_inodes, sizeof(num_of_inodes));
    i += sizeof(num_of_inodes);

    memcpy(&buffer[i], &root_is_data, sizeof(root_is_data));
    i += sizeof(root_is_data);

    memcpy(&buffer[i], &root, sizeof(root));
    i += sizeof(root);

    user_header = &buffer[i];
}


template <class DATA> 
void RTree<DATA>::load_root()
{
    if (root_ptr == NULL)
    // Baum ist nicht leer und root nicht geladen -> root laden 
    {
	if (root_is_data)
	    root_ptr = new RTDataNode<DATA>(this, root);
        else
	    root_ptr = new RTDirNode<DATA>(this, root);
    }    
}

template <class DATA> 
void RTree<DATA>::insert(DATA* d)
{
    int i, j;
    RTNode<DATA> *sn;
    RTDirNode<DATA> *nroot_ptr;
    int nroot;
    DirEntry<DATA> *de;
    R_OVERFLOW split_root;
    DATA *d_cand, *dc;
    float *nmbr;

    // load root into memory
    load_root();

    // no overflow occured until now
    re_level = new bool[root_ptr->level+1];
    for (i = 0; i <= root_ptr->level; i++)
        re_level[i] = false;

    // insert d into re_data_cands as the first entry to insert
    // make a copy of d because it shouldnt be erased later
    dc = new DATA(&dimension);
    *dc = *d;
    re_data_cands.insert(dc);

    j = -1;
    while (re_data_cands.get_num() > 0)
    {
        // first try to insert data, then directory entries
	d_cand = re_data_cands.get_first();
        if (d_cand != NULL)
        {
            // since erase deletes the according data element of the
            // list, we should make a copy of the data before
            // erasing it
            dc = new DATA(&dimension);
            *dc = *d_cand;
            re_data_cands.erase();

	    // start rekursive insert with root
	    split_root = root_ptr->insert(dc, &sn);
	    delete dc;
        }
        else 
	    error("RTree::insert: inconsistent list re_data_cands", true);

	if (split_root == SPLIT)
	// insert has lead to split --> new root-page having to sons
	// root and sn
	{
	    nroot_ptr = new RTDirNode<DATA>(this);
	    nroot_ptr->son_is_data = root_is_data;
	    nroot_ptr->level = root_ptr->level + 1;
	    nroot = nroot_ptr->block;
	    
	    // mbr der alten root page holen und in neue einfuegen
	    de = new DirEntry<DATA>(dimension, this);
	    nmbr = root_ptr->get_mbr();
	    memcpy(de->bounces, nmbr, 2*dimension*sizeof(float));
	    delete nmbr;
	    de->son = root_ptr->block;
	    de->son_ptr = root_ptr;
	    de->son_is_data = root_is_data;
	    de->num_of_data = root_ptr->get_num_of_data();
	    nroot_ptr->enter(de);
	    
	    // mbr der gesplitteten page holen und in neue einfuegen
	    de = new DirEntry<DATA>(dimension, this);
	    nmbr = sn->get_mbr();
	    memcpy(de->bounces, nmbr, 2*dimension*sizeof(float));
	    delete nmbr;
	    de->son = sn->block;
	    de->son_ptr = sn;
	    de->son_is_data = root_is_data;
	    de->num_of_data = sn->get_num_of_data();
	    nroot_ptr->enter(de);
	    
	    root = nroot;
            root_ptr = nroot_ptr;

            // ab dem 1. Split muss die root-page ein Directoryknoten sein
	    root_is_data = false;
        }
        j++;
    }

    // jetzt ist der Baum um eines groesser
    num_of_data++;

    // free lists
    delete re_level;

    // write nodes
/*
    delete root_ptr;
    root_ptr = NULL;
*/
}

template <class DATA> 
DATA * RTree<DATA>::get(int i, int dcopy)
{
    DATA * d;

    // load root into main memory
    load_root();

    d = root_ptr->get(i, dcopy);

    return d;
}

template <class DATA> 
void RTree<DATA>::region(float *mbr)
{
    // load root node into main memory
    load_root();

    root_ptr->region(mbr);
}

template <class DATA>
void RTree<DATA>::NearestNeighborQuery(DATA *QueryPoint, DATA *&Nearest)
{
      float nearest_distanz;

      // load root node into main memory
      load_root();

      nearest_distanz = MAXREAL;

      root_ptr->NearestNeighborSearch(QueryPoint, Nearest, nearest_distanz);
}


template <class DATA>
void RTree<DATA>::NearestNeighborQuery(DATA *QueryPoint,
		SortedLinList<DATA> *&res)
{
      float nearest_distanz;

      // load root node into main memory
      load_root();

      nearest_distanz = MAXREAL;

      root_ptr->NearestNeighborSearch(QueryPoint,res,nearest_distanz);
}

template <class DATA>
void RTree<DATA>::k_NearestNeighborQuery(DATA *QueryPoint, int k, 
		SortedLinList<DATA> *&NeighborList)
{
    int i,l,n;
    float nearest_distanz;
    DATA *p;
    
    nearest_distanz = MAXREAL;
    
    load_root();
    
    // NeighborListListe vorbereiten 
    NeighborList->set_sorting(true);	
    
    for(i = 0; i < k; i++)
    {
	p = new DATA(&dimension);
	for(l = 0; l < dimension; l++)
	    p->data[l] = MAXREAL;
	p->distanz = MAXREAL;	
	NeighborList->insert(p);
    }

    root_ptr->NearestNeighborSearch(QueryPoint,NeighborList,nearest_distanz); 
}		

template <class DATA>
void RTree<DATA>::rangeQuery(DATA *center, float radius,
			SortedLinList<DATA> *res)
{
    load_root();
    root_ptr->rangeQuery(center,radius,res);
}


template <class DATA>
void RTree<DATA>::rangeQuery(float *mbr,
			SortedLinList<DATA> *res)
{
    load_root();

    root_ptr->rangeQuery(mbr,res);
}

template <class DATA>
void RTree<DATA>::rangeQuery(float *mbr,
			LinList<DATA> *res)
{
    load_root();

    root_ptr->rangeQuery(mbr,res);
}

template <class DATA>
void RTree<DATA>::range_nnQuery(float *mbr, SortedLinList<DATA> *res,
			DATA *Nearest)
{
    DATA *center;
    float distanz = MAXREAL;
    bool success = false;
    int i;

    center = new DATA(&dimension);
    for(i = 0; i < dimension; i++)
	center->data[i] = (mbr[2*i] + mbr[2*i+1]) / 2.0; 

    load_root();

    root_ptr->range_nnQuery(mbr,res,center, distanz, Nearest, &success); 
    delete center;
}

template <class DATA>
void RTree<DATA>::SampleQuery(float SampleSizePerCent, LinList<DATA> *res)
{
  int incr = int (100 / SampleSizePerCent);

  for (int i = 0; i < num_of_data; i += incr) {
	DATA *d = get(i);
	res->insert(d);
   }
}

template <class DATA>
void RTree<DATA>::SeedsQuery(LinList<DATA> *samples,
				SortedLinList<DATA> *res)
{
  res->set_sorting(true);
  for ( int i=0; i < samples->get_num(); i++ ) {
	DATA *o = samples->get(i);
	SortedLinList<DATA> *res1 = new SortedLinList<DATA>;
	k_NearestNeighborQuery(o, PTS, res1);
	DATA *p = res1->get(PTS-1);
	o->distanz = p->distanz;
	res->insert(o);
	delete res1;
   }
}

template <class DATA>
void RTree<DATA>::overlapping(float *p, int *nodes_t)
{
    load_root();

    root_ptr->overlapping(p, nodes_t);
}


template <class DATA>
void RTree<DATA>::nodes(int *nodes_a)
{
    load_root();

    root_ptr->nodes(nodes_a);
}


template <class DATA>
void RTree<DATA>::writeinfo(FILE *f)
{
    load_root();
    root_ptr->writeinfo(f);
    fprintf(f,"\n");
}

template <class DATA>
void RTree<DATA>::get_dmbr(float* dmbr,int &length)
{
    load_root();
    root_ptr->get_dmbr(dmbr,length);
}

template <class DATA>
void RTree<DATA>::set_node_weight(int cache_size)
{
    int nodes_tab[20];
    int i, j, rest;
    bool c;

    // calculate number of nodes per level
    for (i = 0; i < 20; i++)
	nodes_tab[i] = 0;

    nodes(nodes_tab);

    rest = cache_size;
    if (cache_size > 0)
	c = true;

    for (i = root_ptr->level; i > 0; i--)
    {
	if (c)
	// cache is not fully filled 
	{
	    if (rest - nodes_tab[i] > 0)
	    // there is enough cache for this level
	    {
		rest -= nodes_tab[i];
		node_weight[i] = 0.0;
	    }
	    else
	    {
		// this is the last level to be chached and not all 
		// nodes of the level are in general in the cache
		node_weight[i] = 1.0 - ((float) rest / (float)nodes_tab[i]);
		rest = 0;
		c = false;
	    }
	}
	else 
	// cache is full
	    node_weight[i] = 1.0;
    }

    // data nodes are not in cache
    node_weight[0] = 1.0;

    printf("cache configuration (%d blocks chache):\n", cache_size);
    for (i = 0; i < 10; i++)
	printf("level %d: %f of %d\n", i, node_weight[i], nodes_tab[i]);
}

#endif // __RTREE

