#include <iostream>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <vector>
#include <queue>
#include <float.h>
#include <math.h>

//for smo
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "svm.h"
#define Malloc(type,n) (type *)malloc((n)*sizeof(type))
//end smo

//for rtree
#include <limits.h>
#include <iostream.h>
#include <fstream.h>
#include <backward/strstream>
#include <unistd.h>
#include <stream.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <queue>
#include <vector>
#include <float.h>
const int CacheSize = 4;
#include "gdbscan.h"
#include "rtree_polygon.h"
#include "rstar_data.h"
#include "fsclinfo.h"
//end rtree

using namespace std;

//for smo
void read_problem(const char *);
void parse_command_line(char **);
struct svm_parameter param;		// set by parse_command_line
struct svm_problem prob;		// set by read_problem
struct svm_model *model;
struct svm_node *x_space;
int cross_validation = 0;
int nr_fold;
//end smo

//for rtree
sClusterInfo<rstar_DATA> *clInfo;
class DensityPt
{
public:
	double _eps;
	rstar_DATA *_dPtr;
	int _dim;
	DensityPt(){}
	DensityPt(double eps, rstar_DATA *dPtr, int dim)
	: _eps(eps), _dPtr(dPtr), _dim(dim) {}
	bool operator< (const DensityPt& d2) const
	{
		return _eps < d2._eps;
	}
	inline void printY(ofstream& fout)
	{
		for(int i = 0;i < _dim;i++)
		{
			fout << _dPtr->data[i] << " ";
		}
		fout << _dPtr->y << "\n";
	}
	inline void printData(ofstream& fout)
	{
		for(int i = 0;i < _dim;i++)
		{
			fout << _dPtr->data[i] << " ";
		}
		fout << "\n";
	}
};
inline double Edist(rstar_DATA *d1, rstar_DATA *d2, int dim)
{
	double sum = 0;
	for(int i = 0;i < dim;i++)
	{
		sum += (d1->data[i] - d2->data[i])*(d1->data[i] - d2->data[i]);
	}
	return sqrt(sum);
}
// ********** rtree error function ********************************************
void error(char *t, bool ex)
{
        cerr << t << endl;
   if (ex) exit(0);
}
//end rtree

void FindAdjMatrix(int ,double ,double *,double ,int ,vector< int >& ,int ,vector< int >& ,bool **);
void FindConnComp(int ,vector< int >& ,bool ** ,int *);
bool checkConnected(int , int , double *, double *, double *, vector< int >& , vector< int >&, double, double);
bool crossConnected(int , int , double *, double *, double *, vector< int >& , vector< int >& , double , double );
int calcAccuracy(int *, int *, int );
inline double kernel(int n, double *z, int k, double gamma)
{
	double val = 0;
	for(int i = 0;i < n;i++)
	{
		val += (z[i]-(model->SV[k])[i].value)*(z[i]-(model->SV[k])[i].value);
	}
	return exp(-val*gamma);
}
int m, M, n, origNumClust, th1, th2;
double g, nu, C, **X;
int *Y, *y;
struct rusage u1;
struct rusage u2;
//Output Variables
double trainAcc, testAcc, OallAcc;
long int smoTime, clustTime, rtreeTime = 0, OallTime;
int le, ls, lsle, actNumClust, denRare;
inline void insertAt(int countNum, float *data, int y)
{
	int tmpIdx = countNum*(n+1);
	prob.x[countNum] = &x_space[tmpIdx];
	prob.y[countNum] = (double)y;
	for(int j = 0;j < n;j++)
	{
		x_space[j + tmpIdx].rowNum = countNum;
		x_space[j + tmpIdx].value = data[j];
		x_space[j + tmpIdx].index = j + 1;
	}
	x_space[tmpIdx + n].index = -1;
	/*for(int j = 0;j < n;j++)
	{
		cout << prob.x[countNum][j].value << " ";
	}
	cout << endl;*/
}


int main(int argc, char **argv)
{
	if((argc != 9) && (argc != 13))
	{
		cerr << "Usage: " << argv[0] << " <m> <M> <n> <k> <gamma> <nu> <TrainFile> <TestFile> (-s or -d <th1> <th2> <0 or 1>)\n";
		return -1;
	}
	m = atoi(argv[1]);
	M = atoi(argv[2]);
	n = atoi(argv[3]);
	origNumClust = atoi(argv[4]);
	g = atof(argv[5]);
	nu = atof(argv[6]);
	bool buildRTree;
	if(argc == 13)
	{
		th1 = atoi(argv[10]);
		th2 = atoi(argv[11]);
		buildRTree = atoi(argv[12]);
	}

	if((argc == 13) && (buildRTree))
	{
		getrusage (RUSAGE_SELF,&u1);
		//Call RTree routines
		//Build RTree
		RTree<rstar_DATA> *rt = new RTree<rstar_DATA>("rmme", 1000, CacheSize, n);
        	float coord;
        	rstar_DATA *d = new rstar_DATA;
        	rstar_DATA *nn;
		FILE *fp = fopen(argv[7],"r");
		if(fp == NULL)
		{
			fprintf(stderr,"can't open input file %s\n", argv[7]);
			return -1;
		}
		int tmpIdx;
		//cout << "Building RTree.....";
		for(int i = 0;i < m;i++)
		{
			fscanf(fp,"%d", &tmpIdx);
			d->y = tmpIdx;
			for(int j = 0;j < n;j++)
			{
				fscanf(fp,"%d:%f", &tmpIdx, &coord);
				d->data[j] = coord;
			}
                	d->weight=1.0;
                	rt->insert(d);
		}
		fclose(fp);
        	delete d;
        	//cout << "done!\n";
		/*int countNum = 0;
		while(d = rt->get(countNum,false))
		{
			for(int j = 0;j < n;j++)
			{
				cout << d->data[j] << " ";
			}
			cout << d->y << "\n";
			countNum++;
		}*/
		if(rt->get_num() < 12)
		{
			cerr << "Too small a problem!\n";
			return -1;
		}
		int const NumN = 10;
		int const NumNd = 4;
		double const nu = 0.75;
		priority_queue<DensityPt> ptQ;
		SortedLinList<rstar_DATA> *nnl;
		nnl = new SortedLinList<rstar_DATA>;
		//Calculation of densities and insertion into heap
		double tmpDist;
		int countNum = 0;
		d = rt->get(countNum,false);
		rt->k_NearestNeighborQuery(d, NumN+1, nnl);
		nn = nnl->get(NumNd+1);
		double *resultant = new double[n];
		double *diff = new double[n];
		double *EpsForAll = new double[rt->get_num()];
		double *ResForAll = new double[rt->get_num()];
		for(int i = 0;i < n;i++)
		{
			resultant[i] = 0;
		}
		for(int i = 1;i <= NumN;i++)
		{
			tmpDist = 0;
			for(int j = 0;j < n;j++)
			{
				diff[j] = (nnl->get(i)->data[j]-d->data[j]);
				tmpDist += diff[j]*diff[j];
			}
			tmpDist = sqrt(tmpDist);
			if(tmpDist != 0)
			{
				for(int j = 0;j < n;j++)
				{
					resultant[j] += diff[j]/tmpDist;
				}
			}
		}
		tmpDist = 0;
		for(int i = 0;i < n;i++)
		{
			tmpDist += resultant[i]*resultant[i];
		}
		tmpDist = sqrt(tmpDist);
		ResForAll[countNum] = tmpDist;
		tmpDist = Edist(d, nn, n);
		EpsForAll[countNum] = tmpDist;
		countNum++;
		//cout << "Starting nn Queries.....";
		delete(nnl);
		while((d = rt->get(countNum, false)) != NULL)
		{
			if(countNum % (rt->get_num()/10) == 0)
			{
				//cout << (double)countNum/rt->get_num()*100 << "% done! ";
			}
			nnl = new SortedLinList<rstar_DATA>;
			rt->k_NearestNeighborQuery(d, NumN+1, nnl);
			nn = nnl->get(NumNd+1);
			for(int i = 0;i < n;i++)
			{
				resultant[i] = 0;
			}
			for(int i = 1;i <= NumN;i++)
			{
				tmpDist = 0;
				for(int j = 0;j < n;j++)
				{
					diff[j] = (nnl->get(i)->data[j]-d->data[j]);
					tmpDist += diff[j]*diff[j];
				}
				tmpDist = sqrt(tmpDist);
				if(tmpDist != 0)
				{
					for(int j = 0;j < n;j++)
					{
						resultant[j] += diff[j]/tmpDist;
					}
				}
			}
			tmpDist = 0;
			for(int i = 0;i < n;i++)
			{
				tmpDist += resultant[i]*resultant[i];
			}
			tmpDist = sqrt(tmpDist);
			ResForAll[countNum] = tmpDist;
			tmpDist = Edist(d, nn, n);
			EpsForAll[countNum] = tmpDist;
			countNum++;
			delete(nnl);
		}
		//cout << "100% done!\n";
		double minRes = DBL_MAX, maxRes = -DBL_MAX, minEps = DBL_MAX, maxEps = -DBL_MAX;
		for(int i = 0;i < countNum;i++)
		{
			if(ResForAll[i] > maxRes)
			{
				maxRes = ResForAll[i];
			}
			if(ResForAll[i] < minRes)
			{
				minRes = ResForAll[i];
			}
			if(EpsForAll[i] > maxEps)
			{
				maxEps = EpsForAll[i];
			}
			if(EpsForAll[i] < minEps)
			{
				minEps = EpsForAll[i];
			}
		}
		//cout << "Starting heap building.....";
		const double minmmaxR = maxRes - minRes, minmmaxE = maxEps - minEps;
		countNum = 0;
		d = rt->get(countNum, false);
		tmpDist = (nu)*((ResForAll[countNum] - minRes)/minmmaxR) + (1-nu)*((EpsForAll[countNum] - minEps)/minmmaxE);
		ptQ.push(DensityPt(tmpDist, d, n));
		countNum++;
		while((d = rt->get(countNum, false)) != NULL)
		{
			tmpDist = (nu)*((ResForAll[countNum] - minRes)/minmmaxR) + (1-nu)*((EpsForAll[countNum] - minEps)/minmmaxE);
			ptQ.push(DensityPt(tmpDist, d, n));
			countNum++;
		}
		//cout << "done\n";
		DensityPt p;
		prob.y = Malloc(double,m);
		prob.x = Malloc(struct svm_node *,m);
		tmpIdx = m*(n+1);
		x_space = Malloc(struct svm_node,tmpIdx);

		//Two possibilities
		if(argv[9][1] == 's')
		{
			//Using the Heap........
			/*DensityPt p;
		        ofstream fout("train.txt",std::ios_base::out);
		        ofstream fout1("test.txt",std::ios_base::out);
		        countNum = 0;
		        cout << " Starting print.....";
		        while(!ptQ.empty())
		        {
		                p = ptQ.top();
		                ptQ.pop();
		                countNum++;
		                if((countNum <= th2) && (countNum > th1))
		                {
		                        p.printY(fout);
		                }
		                else
		                {
		                        p.printY(fout1);
		                }
		        }
		        cout << "done\n";*/
		        countNum = 0;
		        //cout << "Starting reorder.....";
		        while(!ptQ.empty())
		        {
		                p = ptQ.top();
		                ptQ.pop();
				/*tmpIdx = countNum*(n+1);
				prob.x[countNum] = &x_space[tmpIdx];
				prob.y[countNum] = (double)p._dPtr->y;
				for(int j = 0;j < n;j++)
				{
					x_space[j + tmpIdx].rowNum = tmpIdx;
					x_space[j + tmpIdx].value = (double)p._dPtr->data[j];
					x_space[j + tmpIdx].index = j + 1;
				}
				x_space[tmpIdx + n].index = -1;*/
		                if(countNum < th1)
		                {
					tmpIdx = (countNum + th2 - th1)*(n+1);
					prob.x[countNum + th2 - th1] = &x_space[tmpIdx];
					prob.y[countNum + th2 - th1] = (double)p._dPtr->y;
					for(int j = 0;j < n;j++)
					{
						x_space[j + tmpIdx].rowNum = countNum + th2 - th1;
						x_space[j + tmpIdx].value = p._dPtr->data[j];
						x_space[j + tmpIdx].index = j + 1;
					}
					x_space[tmpIdx + n].index = -1;
		                }
		                else
		                {
					if(countNum < th2)
					{
						tmpIdx = (countNum - th1)*(n+1);
						prob.x[countNum - th1] = &x_space[tmpIdx];
						prob.y[countNum - th1] = (double)p._dPtr->y;
						for(int j = 0;j < n;j++)
						{
							x_space[j + tmpIdx].rowNum = countNum - th1;
							x_space[j + tmpIdx].value = p._dPtr->data[j];
							x_space[j + tmpIdx].index = j + 1;
						}
						x_space[tmpIdx + n].index = -1;
					}
					else
					{
						tmpIdx = countNum*(n+1);
						prob.x[countNum] = &x_space[tmpIdx];
						prob.y[countNum] = (double)p._dPtr->y;
						for(int j = 0;j < n;j++)
						{
							x_space[j + tmpIdx].rowNum = countNum;
							x_space[j + tmpIdx].value = p._dPtr->data[j];
							x_space[j + tmpIdx].index = j + 1;
						}
						x_space[tmpIdx + n].index = -1;
					}
		                }
		                countNum++;
		        }
			prob.l = th2 - th1;
		        //cout << "done\n";
		}
		else
		{
			//Two Layer algorithm
			/*ofstream fout("train.txt",std::ios_base::out);
			ofstream fout1("test.txt",std::ios_base::out);*/
			countNum = 0;
			bool *printed = new bool[rt->get_num()];
			for(int i = 0;i < rt->get_num();i++)
			{
				printed[i] = false;
			}
			vector< DensityPt > borderPts(th2-th1);
			vector< DensityPt > corePts;
			while(!ptQ.empty())
			{
				p = ptQ.top();
				ptQ.pop();
				countNum++;
				p._dPtr->rank = countNum;
				if((countNum <= th2) && (countNum > th1))
				{
					borderPts[countNum-th1-1] = p;
					insertAt(countNum-th1-1, p._dPtr->data, p._dPtr->y);
					printed[countNum-1] = true;
				}
				else
				{
					corePts.push_back(p);
				}
			}
			countNum = 0;
			for(int i = 0;i < th2-th1;i++)
			{
				nnl = new SortedLinList<rstar_DATA>;
				rt->k_NearestNeighborQuery(borderPts[i]._dPtr, 4, nnl);
				nn = nnl->get(1);
				if((nn->rank > th2) && (printed[nn->rank - 1] == false))
				{
					insertAt(countNum+th2-th1, nn->data, nn->y);
					countNum++;
					printed[nn->rank - 1] = true;
				}
				else
				{
					nn = nnl->get(2);
					if((nn->rank > th2) && (printed[nn->rank - 1] == false))
					{
						insertAt(countNum+th2-th1, nn->data, nn->y);
						countNum++;
						printed[nn->rank - 1] = true;
					}
					else
					{
						nn = nnl->get(3);
						if((nn->rank > th2) && (printed[nn->rank - 1] == false))
						{
							insertAt(countNum+th2-th1, nn->data, nn->y);
							countNum++;
							printed[nn->rank - 1] = true;
						}
						/*else
						{
							cout << "Border Pt " << i + 1 << "'s nn not found\n";
						}*/
					}
				}
				delete(nnl);
			}
			countNum += th2 - th1;
			for(int i = 0, j = 0;i < rt->get_num()-(th2-th1);i++)
			{
				if(printed[corePts[i]._dPtr->rank - 1] == false)
				{
					insertAt(countNum+j, corePts[i]._dPtr->data, corePts[i]._dPtr->y);
					j++;
					printed[corePts[i]._dPtr->rank - 1] = true;
				}
			}
			//cout << "done!!!\n";
			prob.l = 2*(th2 - th1);
			
			delete[] printed;
		}
		
		//Finishing Rtree
		delete rt;
		delete[] resultant;
		delete[] diff;
		delete[] EpsForAll;
		delete[] ResForAll;
		getrusage (RUSAGE_SELF,&u2);
		rtreeTime = u2.ru_utime.tv_sec - u1.ru_utime.tv_sec;
		//cout << "done!!!!\n";
		ofstream frout("rtree.out");
		for(int i = 0;i < m;i++)
		{
			frout << prob.y[i];
			for(int j = 0;j < n;j++)
			{
				frout << " " << j+1 << ":" << (prob.x[i])[j].value;
			}
			frout << "\n";
		}
		frout.close();
	}
	else
	{
		if(argc == 9)
		{
			read_problem(argv[7]);
		}
		else
		{
			read_problem("rtree.out");
			prob.l = 2*(th2-th1);
		}
	}
	parse_command_line(argv);
	X = new double*[M];
	for(int i = 0;i < M;i++)
	{
		X[i] = new double[n];
	}
	Y = new int[M];
	FILE *fp = fopen(argv[8],"r");
	if(fp == NULL)
	{
		fprintf(stderr,"can't open input file %s\n", argv[8]);
		return -1;
	}
	int tmpIdx;
	for(int i = 0;i < M;i++)
	{
		fscanf(fp,"%d", &Y[i]);
		for(int j = 0;j < n;j++)
		{
			fscanf(fp,"%d:%lf", &tmpIdx, &X[i][j]);
		}
	}
	fclose(fp);

	/*for(int i = 0;i < prob.l;i++)
	{
		for(int j = 0;j < n;j++)
		{
			cout << (prob.x[i])[j].value << " ";
		}
		cout << prob.y[i] << "\n";
	}
	return 0;*/
	/*for(int i = 0;i < M;i++)
	{
		for(int j = 0;j < n;j++)
		{
			cout << X[i][j] << " ";
		}
		cout << Y[i] << "\n";
	}*/

	//Call SMO routines
	const char *error_msg = svm_check_parameter(&prob,&param);
	if(error_msg)
	{
		fprintf(stderr,"Error: %s\n",error_msg);
		exit(1);
	}
//	cout << param.gamma << " " << param.nu << endl;
	getrusage (RUSAGE_SELF,&u1);
	model = svm_train(&prob,&param);
	getrusage (RUSAGE_SELF,&u2);
	smoTime = u2.ru_utime.tv_sec - u1.ru_utime.tv_sec;
	
	//Call Clustering routines
	C = 1.0/(nu*m);
	double *alpha = model->sv_coef[0], rho = C*model->rho[0];
	svm_node **SV = model->SV;
	lsle = model->l;
	/*for(int i = 0;i < lsle;i++)
	{
		cout << SV[i][0].rowNum << " " << alpha[i] << "\n";
	}*/
	getrusage (RUSAGE_SELF,&u1);
	ls = 0;
	le = 0;
	vector< int > inds,inde;
	vector< int > globalInds, globalInde;
	for(int i = 0;i < lsle;i++)
	{
		if(alpha[i] == 1)
		{
			++le;
			inde.push_back(i);
			globalInde.push_back(SV[i][0].rowNum);
		}
		else
		{
			++ls;
			inds.push_back(i);
			globalInds.push_back(SV[i][0].rowNum);
		}
		alpha[i] = alpha[i]*C;
	}
	
	bool **adjMatrix = new bool*[ls];
        for(int i = 0;i < ls;i++)
        {
                adjMatrix[i] = new bool[ls];
        }
        for(int i = 0;i < ls;i++)
        {
                for(int j = 0;j < ls;j++)
                {
                        adjMatrix[i][j] = false;
                }
        }
        FindAdjMatrix(n, g, alpha, rho, ls, inds, le, inde, adjMatrix);
        /*for(int i = 0;i < ls;i++)
        {
                for(int j = 0;j < ls;j++)
                {
                        cout << adjMatrix[i][j] << " ";
                }
		cout << "\n";
        }*/
        int *clustAssign = new int[m];
	for(int i = 0;i < m;i++)
	{
		clustAssign[i] = 0;
	}
        FindConnComp(ls, inds, adjMatrix, clustAssign);
	/*for(int i = 0;i < m;i++)
	{
		cout << clustAssign[i] << "\n";
	}*/
	
	//NNC for training set
	int currSV = 0, nn, lnn, currEV = 0;
        double dist, minDist;
	double *diff = new double[n];
	double *z = new double[n];
        for(int i = 0;i < m;i++)
        {
                if(i != globalInds[currSV])
                {
                        //Calc nearest SV
                        minDist = DBL_MAX;
                        for(int j = 0;j < ls;j++)
                        {
                                dist = 0;
                                for(int k = 0;k < n;k++)
                                {
                                        dist += ((prob.x[i])[k].value - (model->SV[inds[j]])[k].value)*((prob.x[i])[k].value - (model->SV[inds[j]])[k].value);
                                }
                                if(dist < minDist)
                                {
                                        minDist = dist;
                                        nn = globalInds[j];
					lnn = j;
                                }
                        }
                       	clustAssign[i] = clustAssign[nn];
			//if nn SV is valid
			if((le > 0) && (i == globalInde[currEV]))
			{
				if(currEV != (le - 1))
				{
					++currEV;
				}
			}
			else
			{
				bool valid = checkConnected(i,nn,diff,z,alpha,inds,inde,rho,g);
				if(!valid)
				{
					for(int ii = 0;ii < ls;ii++)
					{
						if(ii != lnn)
						{
							valid = checkConnected(i,globalInds[ii],diff,z,alpha,inds,inde,rho,g);
							if(valid)
							{
	                        				clustAssign[i] = clustAssign[globalInds[ii]];
								break;
							}
						}
					}
				}
			}
                }
                else
                {
			if(currSV != (ls - 1))
			{
                        	currSV++;
			}
                }
       	}
	getrusage (RUSAGE_SELF,&u2);
	clustTime = u2.ru_utime.tv_sec - u1.ru_utime.tv_sec;
	//NNC for test set
        int *newClustAssign = new int[M];
	for(int i = 0;i < M;i++)
	{
		newClustAssign[i] = 0;
	}
	bool outClustPt;
        for(int i = 0;i < M;i++)
        {
		//evaluate decision function
		dist = 0;
		for(int k = 0;k < ls;k++)
		{
			dist += kernel(n, X[i], inds[k], g)*alpha[inds[k]];
		}
		for(int k = 0;k < le;k++)
		{
			dist += kernel(n, X[i], inde[k], g)*alpha[inde[k]];
		}
		dist -= rho;
		if(dist < 0)
		{
			outClustPt = true;
		}
		else
		{
			outClustPt = false;
		}
		//Calc nearest SV
		minDist = DBL_MAX;
		for(int j = 0;j < ls;j++)
		{
			dist = 0;
			for(int k = 0;k < n;k++)
			{
				dist += (X[i][k] - (model->SV[inds[j]])[k].value)*(X[i][k] - (model->SV[inds[j]])[k].value);
			}
			if(dist < minDist)
			{
				minDist = dist;
				nn = globalInds[j];
				lnn = j;
			}
		}
		newClustAssign[i] = clustAssign[nn];
		//if nn SV is valid
		if(!outClustPt)
		{
			bool valid = crossConnected(i,nn,diff,z,alpha,inds,inde,rho,g);
			if(!valid)
			{
				for(int ii = 0;ii < ls;ii++)
				{
					if(ii != lnn)
					{
						valid = crossConnected(i,globalInds[ii],diff,z,alpha,inds,inde,rho,g);
						if(valid)
						{
	                       				newClustAssign[i] = clustAssign[globalInds[ii]];
							break;
						}
					}
				}
			}
		}
       	}
	delete[] diff;
	delete[] z;
	/*for(int i = 0;i < m;i++)
	{
		cout << clustAssign[i] << "\n";
	}*/
	/*for(int i = 0;i < M;i++)
	{
		cout << newClustAssign[i] << "\n";
	}*/
	actNumClust = -1;
	y = new int[m];
	for(int i = 0;i < m;i++)
	{
		y[i] = (int)floor(prob.y[i]);
		if(actNumClust < clustAssign[i])
		{
			actNumClust = clustAssign[i];
		}
	}
	svm_destroy_model(model);
	svm_destroy_param(&param);
	trainAcc = calcAccuracy(y, clustAssign, m);
	testAcc = calcAccuracy(Y, newClustAssign, M);
	OallAcc = ((double) (trainAcc + testAcc) / (m + M))*100;
	trainAcc = ((double) trainAcc / m)*100;
	testAcc = ((double) testAcc / M)*100;
	OallTime = smoTime + clustTime + rtreeTime;
	vector< int > freqClust(actNumClust);
	for(int i = 0;i < actNumClust;i++)
	{
		freqClust[i] = 0;
	}
	for(int i = 0;i < m;i++)
	{
		freqClust[clustAssign[i]-1]++;
	}
	denRare = INT_MAX;
	for(int i = 0;i < actNumClust;i++)
	{
		if(freqClust[i] < denRare)
		{
			denRare = freqClust[i];
		}
	}
	//Output
	cout << g << " " << nu << " " << trainAcc << " " << testAcc << " " << OallAcc << " " << smoTime << " " << clustTime << " " << rtreeTime << " " << OallTime << " " << le << " " << ls << " " << lsle << " " << actNumClust << " " << denRare << endl;
	//cerr << g << " " << nu << " " << trainAcc << " " << testAcc << " " << OallAcc << " " << smoTime << " " << clustTime << " " << rtreeTime << " " << OallTime << " " << le << " " << ls << " " << lsle << " " << actNumClust << " " << denRare << endl;
	ofstream ofs("try.txt");
	for(int i = 0;i < m;i++)
	{
		ofs << clustAssign[i] << " " << y[i] <<"\n";
	}
	//Fininshing.....
	for(int i = 0;i < M;i++)
	{
		delete[] X[i];
	}
	delete[] X;
	delete[] Y;
	delete[] y;
	free(prob.y);
	free(prob.x);
	free(x_space);
	delete[] clustAssign;
	delete[] newClustAssign;
        for(int i = 0;i < ls;i++)
        {
                delete[] adjMatrix[i];
        }
	delete[] adjMatrix;

	return 0;
}

void read_problem(const char *filename)
{
	int elements, max_index, i, j;
	FILE *fp = fopen(filename,"r");
	
	if(fp == NULL)
	{
		fprintf(stderr,"can't open input file %s\n",filename);
		exit(1);
	}

	prob.l = 0;
	elements = 0;
	while(1)
	{
		int c = fgetc(fp);
		switch(c)
		{
			case '\n':
				++prob.l;
				// fall through,
				// count the '-1' element
			case ':':
				++elements;
				break;
			case EOF:
				goto out;
			default:
				;
		}
	}
out:
	rewind(fp);
	prob.y = Malloc(double,prob.l);
	prob.x = Malloc(struct svm_node *,prob.l);
	x_space = Malloc(struct svm_node,elements);

	max_index = 0;
	j=0;
	for(i=0;i<prob.l;i++)
	{
		double label;
		prob.x[i] = &x_space[j];
		fscanf(fp,"%lf",&label);
		prob.y[i] = label;
		while(1)
		{
			int c;
			do {
				c = getc(fp);
				if(c=='\n') goto out2;
			} while(isspace(c));
			ungetc(c,fp);
			fscanf(fp,"%d:%lf",&(x_space[j].index),&(x_space[j].value));
			x_space[j].rowNum = i;
			++j;
		}
out2:
		if(j>=1 && x_space[j-1].index > max_index)
			max_index = x_space[j-1].index;
		x_space[j++].index = -1;
	}

	if(param.gamma == 0)
		param.gamma = 1.0/max_index;

	fclose(fp);
}

void parse_command_line(char **argv)
{
	// default values
	param.svm_type = ONE_CLASS;
	param.kernel_type = RBF;
	param.degree = 3;
	param.gamma = atof(argv[5]);	// 1/k
	param.coef0 = 0;
	param.nu = atof(argv[6]);
	param.cache_size = 40;
	param.C = 1;
	param.eps = 1e-3;
	param.p = 0.1;
	param.shrinking = 1;
	param.probability = 0;
	param.nr_weight = 0;
	param.weight_label = NULL;
	param.weight = NULL;
}

void FindAdjMatrix(int n, double gamma, double *alpha, double rho, int ls, vector< int >& inds, int le, vector< int >& inde, bool **adjMatrix)
{
	bool adj_flag;
	double *diff = new double[n];
	double *z = new double[n];
	double d;
	for(int i = 0;i < ls;i++)
	{
		/*if(i % (ls/10) == 0)
		{
			cout << ((double)i/ls)*100 << "\% done.\n";
		}*/
		for(int j = 0;j < ls;j++)
		{
			if(j < i)
			{
				if(adjMatrix[i][j])
				{
					for(int k = j+1;k < ls;k++)
					{
						adjMatrix[i][k] = (adjMatrix[i][k] || adjMatrix[j][k]);
					}
				}
			}
			else
			{
				if((i < j) && (!adjMatrix[i][j]))
				{
					//cout << "i: " << i << " j: " << j << " ali: " << alpha[inds[i]] << " alj: " << alpha[inds[j]] << endl;
					//cout << "i: " << i << " j: " << j << " int: " << interval << " k: " << k << " ker: " << kernel(n, z, inds[k], gamma) << " z0: " << z[0] << " z1: " << z[1] << " k0: " << (model->SV[inds[k]])[0].value << " k1: " << (model->SV[inds[k]])[1].value << endl;
					//goes over 10 points in the interval between these 2 Sample points
					adj_flag = true; //unless a point on the path exits the shpere - the points are adj
					for(int k = 0;k < n;k++)
					{
						diff[k] =  (model->SV[inds[j]])[k].value - (model->SV[inds[i]])[k].value;
					}
					for(double interval = 0.1;interval < 0.95;interval+=0.1)
					{
						for(int k = 0;k < n;k++)
						{
							z[k] = (model->SV[inds[i]])[k].value + interval * diff[k];
						}
						//calculates the sub-point distance from the sphere's center
						d = 0;
						for(int k = 0;k < ls;k++)
						{
							d += kernel(n, z, inds[k], gamma)*alpha[inds[k]];
						}
						for(int k = 0;k < le;k++)
						{
							d += kernel(n, z, inde[k], gamma)*alpha[inde[k]];
						}
						d -= rho;
						//cout << "i: " << i << " j: " << j << " int: " << interval << " ker: " << kernel(n, z, inds[0], gamma) << " z0: " << z[0] << " z1: " << z[1] << " SV0: " << (model->SV[inds[0]])[0].value << " SV1: " << (model->SV[inds[0]])[1].value << endl;
						//cout << i << " " << j << " " << interval << " " << d << endl;
						/*if((i == 0) && (j == 3))
						{
							cout << d << "\n";
						}*/
						if(d < 0)
						{
							/*if(fabs(d) < 0.00001)
							{
								cout << d << "\n";
							}*/
							adj_flag = false;
							break;
						}
					}
					if(adj_flag)
					{
						adjMatrix[i][j] = true;
						adjMatrix[j][i] = true;
					}
				}
				else
				{
					adjMatrix[i][j] = true;
				}
			}
		}
	}
	delete[] diff;
	delete[] z;
}

void FindConnComp(int ls, vector< int >& inds,bool **adjMatrix,int *clustAssign)
{
        bool *visited = new bool[ls];
        for(int i = 0;i < ls;i++)
        {
                visited[i] = false;
        }
                                                                                                                            
        queue< int > Q;
        int currNode, currClust = 1;
        for(int i = 0;i < ls;i++)
        {
                if(!visited[i])
                {
                        visited[i] = true;
                        Q.push(i);
                        while(!Q.empty())
                        {
                                currNode = Q.front();
                                Q.pop();
                                for(int j = 0;j < ls;j++)
                                {
                                        if((adjMatrix[currNode][j]) && (!visited[j]))
                                        {
                                                visited[j] = true;
                                                Q.push(j);
                                        }
                                }
                                clustAssign[(model->SV[inds[currNode]])[0].rowNum] = currClust;
                        }
                        currClust++;
                }
        }
        delete[] visited;
}

bool checkConnected(int i, int nn, double *diff, double *z, double *alpha, vector< int >& inds, vector< int >& inde, double rho, double gamma)
{
	bool valid = true;
	double d;
	for(int k = 0;k < n;k++)
	{
		diff[k] =  (prob.x[i])[k].value - (prob.x[nn])[k].value;
	}
	for(double interval = 0.1;interval < 0.95;interval+=0.1)
	{
		for(int k = 0;k < n;k++)
		{
			z[k] = (prob.x[nn])[k].value + interval * diff[k];
		}
		//calculates the sub-point distance from the sphere's center
		d = 0;
		for(int k = 0;k < ls;k++)
		{
			d += kernel(n, z, inds[k], gamma)*alpha[inds[k]];
		}
		for(int k = 0;k < le;k++)
		{
			d += kernel(n, z, inde[k], gamma)*alpha[inde[k]];
		}
		d -= rho;
		if(d < 0)
		{
			valid = false;
			break;
		}
	}
	return valid;
}

bool crossConnected(int i, int nn, double *diff, double *z, double *alpha, vector< int >& inds, vector< int >& inde, double rho, double gamma)
{
	bool valid = true;
	double d;
	for(int k = 0;k < n;k++)
	{
		diff[k] =  X[i][k] - (prob.x[nn])[k].value;
	}
	for(double interval = 0.1;interval < 0.95;interval+=0.1)
	{
		for(int k = 0;k < n;k++)
		{
			z[k] = (prob.x[nn])[k].value + interval * diff[k];
		}
		//calculates the sub-point distance from the sphere's center
		d = 0;
		for(int k = 0;k < ls;k++)
		{
			d += kernel(n, z, inds[k], gamma)*alpha[inds[k]];
		}
		for(int k = 0;k < le;k++)
		{
			d += kernel(n, z, inde[k], gamma)*alpha[inde[k]];
		}
		d -= rho;
		if(d < 0)
		{
			valid = false;
			break;
		}
	}
	return valid;
}

int calcAccuracy(int *y, int *Y, int m)
{
	//Accuracy Calculation
	int maxy = -1;
	for(int i = 0;i < m;i++)
	{
		if(maxy < y[i])
		{
			maxy = y[i];
		}
	}
	vector< int > origClust(maxy);
	for(int i = 0;i < maxy;i++)
	{
		origClust[i] = i;
	}
	int maxY = -1;
	for(int i = 0;i < m;i++)
	{
		if(maxY < Y[i])
		{
			maxY = Y[i];
		}
	}
	vector< int > newClust(maxY);
	for(int i = 0;i < maxY;i++)
	{
		newClust[i] = i;
	}
	int **confusion = new int*[maxy];
	for(int i = 0;i < maxy;i++)
	{
		confusion[i] = new int[maxY];
	}
	for(int i = 0;i < maxy;i++)
	{
		for(int j = 0;j < maxY;j++)
		{
			confusion[i][j] = 0;
		}
	}
    	for(int i = 0;i < m;i++)
	{
		confusion[y[i]-1][Y[i]-1]++;
	}
    
    	int currMax, maxi, maxj, numCorrClust = 0;
	vector< int >::iterator pDel;
	/*for(int i = 0;i < maxy;i++)
	{
		for(int j = 0;j < maxY;j++)
		{
			cout << confusion[i][j] << " ";
		}
		cout << endl;
	}*/
	while((newClust.size() != 0) && (origClust.size() != 0))
	{
		currMax = -1;
		for(int i = 0;i < origClust.size();i++)
		{
			//cout << origClust[i] << " ";
			for(int j = 0;j < newClust.size();j++)
			{
				if(currMax < confusion[origClust[i]][newClust[j]])
				{
					currMax = confusion[origClust[i]][newClust[j]];
					maxi = i;
					maxj = j;
				}
			}
		}
		/*cout << endl;
		for(int j = 0;j < newClust.size();j++)
		{
			cout << newClust[j] << " ";
		}
		cout << endl;*/
		numCorrClust += currMax;
		pDel = newClust.begin();
		pDel += maxj;
		newClust.erase(pDel);
		pDel = origClust.begin();
		pDel += maxi;
		origClust.erase(pDel);
	}
	for(int i = 0;i < maxy;i++)
	{
		delete[] confusion[i];
	}
	delete[] confusion;
	return numCorrClust;
}
